"""High-level Object-oriented API for the different types of packets
(FixedLength and VariableLength) supported by the package.
"""

import csv
import os
import warnings

import numpy as np

from .converters import Converter
from .decode import _decode_fixed_length, _decode_variable_length
from .packet_fields import PacketField, PacketArray


__author__ = "Daniel da Silva <mail@danieldasilva.org>"


class _BasePacket:
    """Base class of FixedLength and VariableLength. Not to be instantiated
    directly.
    """

    def _init(self, fields):
        """
        Parameters
        ----------
        fields : list of `ccsdspy.PacketField`
            Layout of packet fields contained in the definition.
        """
        # List of PacketField instances
        self._fields = fields[:]

        # Dictionary mapping input name to tuple (output_name: str, Converter instance)
        self._converters = {}

    @classmethod
    def from_file(cls, file):
        """
        Parameters
        ----------
        file : str
           Path to file on the local file system that defines the packet fields.
           Currently only supports csv files.
           See :download:`simple_csv_3col.csv <../../ccsdspy/tests/data/packet_def/simple_csv_3col.csv>`  # noqa: E501
           and :download:`simple_csv_4col.csv <../../ccsdspy/tests/data/packet_def/simple_csv_4col.csv>`  # noqa: E501

        Returns
        -------
        An instance of FixedLength.
        """
        file_extension = os.path.splitext(file)
        if file_extension[1] == ".csv":
            fields = _get_fields_csv_file(file)
        else:
            raise ValueError(f"File type {file_extension[1]} not supported.")

        return cls(fields)

    def add_converted_field(self, input_field_name, output_field_name, converter):
        """Add a converted field to the packet definition, used to apply
        post-processing transformations of decoded fields.

        Parameters
        ----------
        input_field_name : str or list/tuple
           Name of input field, or list/tuple of names of fields. There must be field(s)
           which exists in the packet definition corresponding to these name(s).
        output_field_name : str
           Name of output field. When the packet is decoded using `pkt.load()`,
           a new field named this will be present in the output dictionary.
        converter : instance of subclass of `:py:class:~ccsdspy.converters.Converter`
           A converter object to apply post-processing conversions, such as
           calibration curves or value replacement. Converter objects
           can be found in`:py:mod:~ccsdspy.converters`.

        Raises
        ------
        TypeError
           If one of the arguments is not of the correct type.
        ValueError
           The provided `input_field_name` is not present in the packet definition
        """
        if not isinstance(output_field_name, str):
            raise TypeError("output_field_name must be a str")
        if not isinstance(converter, Converter):
            raise TypeError("converter must be an instance of a Converter subclass")

        # Get tuple of input field names for storing; this handles the input_field_name
        # argument being either a str, or list/tuple
        if isinstance(input_field_name, str):
            input_field_names = (input_field_name,)
        elif isinstance(input_field_name, (list, tuple)):
            input_field_names = tuple(input_field_name)
        else:
            raise TypeError("input_field_name must be either str, list, or tuple")

        del input_field_name  # don't use the variable again in this function

        # Check that each of the input field names exists in the packet, and report
        # the missing fields if not
        fields_in_packet_set = {field._name for field in self._fields}
        input_field_names_set = set(input_field_names)
        all_fields_present = input_field_names_set <= fields_in_packet_set  # subset

        if not all_fields_present:
            missing_fields = input_field_names_set - fields_in_packet_set  # set op A \ B
            raise ValueError(
                "Some fields specified as inputs to converters were missing: "
                f"{sorted(missing_fields)}"
            )

        self._converters[input_field_names] = (output_field_name, converter)


class FixedLength(_BasePacket):
    """Define a fixed length packet to decode binary data.

    Fixed length packets correspond to packets that are the same length and
    layout every time. A common example of this is housekeeping or status
    messages.
    """

    def __init__(self, fields):
        """
        Parameters
        ----------
        fields : list of :py:class:`~ccsdspy.PacketField` or :py:class:`~ccsdspy.PacketArray`
            Layout of packet fields contained in the definition.

        Raises
        ------
        ValueError
            one or more of the arguments are invalid
        """
        if any(isinstance(field._array_shape, str) for field in fields):
            raise ValueError(
                "The FixedLength class does not support variable fields. "
                "Instead, use the VariableLength class."
            )

        self._init(fields)

    def load(self, file, include_primary_header=False):
        """Decode a file-like object containing a sequence of these packets.

        Parameters
        ----------
        file : str
           Path to file on the local file system, or file-like object
        include_primary_header : bool
           If True, provides the primary header in the output

        Returns
        -------
        field_arrays : dict, string to NumPy array
            dictionary mapping field names to NumPy arrays, with key order matching
            the order of fields in the packet.

        Warns
        -----
        UserWarning
            If the ccsds sequence count is not in order
        UserWarning
            If the ccsds sequence count is missing packets
        UserWarning
            If there are more than one APID
        """
        packet_arrays = _load(
            file,
            self._fields,
            self._converters,
            "fixed_length",
            include_primary_header=True,
        )

        # inspect the primary header and issue warning if appropriate
        _inspect_primary_header_fields(packet_arrays)

        if not include_primary_header:
            _delete_primary_header_fields(packet_arrays)

        return packet_arrays


class VariableLength(_BasePacket):
    """Define a variable length packet to decode binary data.

    Variable length packets are packets which have a different length each
    time.  Variable length fields are defined as `~ccsdspy.PacketArray` fields
    where `array_shape="expand"` (causing the field to grow to fill the packet) or
    `array_shape="other_field"` (causes the field named `other_field` to set the number
    of elements in this array).

    Please note that while this class is able to parse fixed length packets, it
    is much slower. Use the :py:class:`~ccsdspy.FixedLength` class instead.

    Rules for variable length packets:
      * Do only specify a `~ccsdspy.PacketArray` with the `array_shape="other_field"`
        when `other_field` precedes it in the packet definition
      * Do not provide more than one expanding `~ccsdspy.PacketArray` with `array_shape="expand"`
      * Do not specify the primary header fields manually
      * Do not specify explicit bit_offsets (they will be computed automatically)
    """

    def __init__(self, fields):
        """
        Parameters
        ----------
        fields : list of :py:class:`~ccsdspy.PacketField` or :py:class:`~ccsdspy.PacketArray`
            Layout of packet fields contained in the definition. No more than
            one field should have array_shape="expand". The field must have no
            bit_offset's. Do not include the primary header fields.

        Raises
        ------
        ValueError
            one or more of the arguments are invalid, or do not follow the
            specified rules.
        """
        # Check there is only one expanding field in the packet definition
        expand_arrays = [
            field
            for field in fields
            if isinstance(field, PacketArray) and field._array_shape == "expand"
        ]

        if len(expand_arrays) > 1:
            raise ValueError(
                "The VariableLength class does not support more than one field "
                "with array_shape='expand', as the decoding process becomes "
                "ambiguous."
            )

        # Check variable fields with their sizes set by other fields only do so when
        # the previous field precedes it
        field_names = [field._name for field in fields]

        for i, field in enumerate(fields):
            if (
                isinstance(field, PacketArray)
                and isinstance(field._array_shape, str)
                and field._array_shape != "expand"
                and field._array_shape not in field_names[:i]
            ):
                raise ValueError(
                    "The VariableLength class requires that variable fields with "
                    "their sizes set by other fields only do so when the "
                    "previous field precedes it."
                )

        # Check that bit offsets are not set
        if not all(field._bit_offset is None for field in fields):
            raise ValueError(
                "The VariableLength class does not support explicit bit "
                "offsets. You must specify the entire packet so they can be "
                "determined automatically."
            )

        self._init(fields)

    def load(self, file, include_primary_header=False):
        """Decode a file-like object containing a sequence of these packets.

        Parameters
        ----------
        file : str
           Path to file on the local file system, or file-like object
        include_primary_header : bool
           If True, provides the primary header in the output

        Returns
        -------
        field_arrays : dict, string to NumPy array
            dictionary mapping field names to NumPy arrays, with key order matching
            the order of fields in the packet.

        Warns
        -----
        UserWarning
            If the ccsds sequence count is not in order
        UserWarning
            If the ccsds sequence count is missing packets
        UserWarning
            If there are more than one APID
        """
        # The variable length decoder requires the full packet definition, so if
        # they didn't want the primary header fields, we parse for them and then
        # remove them after.
        packet_arrays = _load(
            file, self._fields, self._converters, "variable_length", include_primary_header=True
        )

        # inspect the primary header and issue warning if appropriate
        _inspect_primary_header_fields(packet_arrays)

        if not include_primary_header:
            _delete_primary_header_fields(packet_arrays)

        return packet_arrays


def _inspect_primary_header_fields(packet_arrays):
    """Inspects the primary header fields.

    Checks for the following issues
    * all apids are the same
    * sequence count is not missing any values
    * sequence count is in order

    Parameters
    -----------
    packet_arrays
        dictionary mapping field names to NumPy arrays, with key order matching
        the order fields in the packet. Modified in place
    """
    seq_counts = packet_arrays["CCSDS_SEQUENCE_COUNT"]
    start, end = seq_counts[0], seq_counts[-1]
    missing_elements = sorted(set(range(start, end + 1)).difference(seq_counts))
    if len(missing_elements) != 0:
        warnings.warn(f"Missing packets found {missing_elements}.", UserWarning)

    if not np.all(seq_counts == np.sort(seq_counts)):
        warnings.warn("Sequence count are out of order.", UserWarning)

    individual_ap_ids = set(packet_arrays["CCSDS_APID"])
    if len(individual_ap_ids) != 1:
        warnings.warn(f"Found multiple AP IDs {individual_ap_ids}.", UserWarning)

    return None


def _delete_primary_header_fields(packet_arrays):
    """Modifies in place the packet arrays dictionary to delete primary
    header fields.

    Parameters
    -----------
    packet_arrays
        dictionary mapping field names to NumPy arrays, with key order matching
        the order fields in the packet. Modified in place
    """
    header_fields = _prepend_primary_header_fields([])

    for header_field in header_fields:
        del packet_arrays[header_field._name]


def _expand_array_fields(existing_fields):
    """Expand arrays into multiple fields, one for each element.

    Returns a new list of fields as well as a data structure which can be used
    to reverse this process. See the `_unexpand_field_arrays()` function to reverse
    this process.

    Parameters
    ----------
    existing_fields : list of `ccsdspy.PacketField`
      Layout of packet fields contained in the definition, with PacketArray

    Returns
    -------
    return_fields : list of `ccsdspy.PacketField`
      Layout of packet fields contained in the definition, without PacketArray's
    expand_history : dict
      Dictionary mapping array name with shape/data-type and field expansions
    """
    return_fields = []
    expand_history = {}

    for existing_field in existing_fields:
        if existing_field._field_type != "array" or isinstance(existing_field._array_shape, str):
            return_fields.append(existing_field)
            continue

        array_shape = existing_field._array_shape
        array_order = existing_field._array_order

        index_vecs = [np.arange(dim) for dim in array_shape]
        index_grids = np.meshgrid(*index_vecs, indexing="ij")
        indices_flat = [index_grid.flatten(order=array_order) for index_grid in index_grids]

        expand_history[existing_field._name] = {
            "shape": array_shape,
            "data_type": existing_field._data_type,
            "fields": {},
        }

        for i, indices in enumerate(zip(*indices_flat)):
            name = f"{existing_field._name}[{','.join(map(str,indices))}]"
            if existing_field._bit_offset is None:
                bit_offset = None
            else:
                bit_offset = existing_field._bit_offset + i * existing_field._bit_length

            return_field = PacketField(
                name=name,
                data_type=existing_field._data_type,
                bit_length=existing_field._bit_length,
                bit_offset=bit_offset,
                byte_order=existing_field._byte_order,
            )

            expand_history[existing_field._name]["fields"][name] = indices
            return_fields.append(return_field)

    return return_fields, expand_history


def _unexpand_field_arrays(field_arrays, expand_history):
    """Reverse the array expansion process from `_expand_array_fields`.

    Parameters
    ----------
    field_arrays : dict, str to numpy array
      Dictionary mapping field names to NumPy arrays, with key order matching
      the order fields in the packet. Has a key for each array element.
    expand_history : dict
      Dictionary mapping array name with shape/data-type and field expansions

    Returns
    -------
    return_field_arrays : dict, str to array
      Dictionary mapping field names to NumPy arrays, with key order matching
      the order fields in the packet. Has keys mapping to full arrays.
    """
    npackets = list(field_arrays.values())[0].shape[0]
    return_field_arrays = field_arrays.copy()

    for array_name, array_details in expand_history.items():
        array_shape = (npackets,) + array_details["shape"]
        array_dtype = field_arrays[list(array_details["fields"].keys())[0]].dtype
        array = np.zeros(array_shape, dtype=array_dtype)

        for element_name, indices in array_details["fields"].items():
            array.__setitem__((slice(None),) + indices, field_arrays[element_name])
            del return_field_arrays[element_name]

        return_field_arrays[array_name] = array

    return return_field_arrays


def _prepend_primary_header_fields(existing_fields):
    """Helper function that prepends primary header fields to a list of packet
    fields, to support load(include_primary_header=True)

    Parameters
    ----------
    existing_fields : list of `ccsdspy.PacketField`
      Non-primary header fields defined by the packet.

    Returns
    -------
    New list of fields with the primary header fields prepended.
    """
    return_fields = [
        PacketField(
            name="CCSDS_VERSION_NUMBER",
            data_type="uint",
            bit_length=3,
            bit_offset=0,
        ),
        PacketField(
            name="CCSDS_PACKET_TYPE",
            data_type="uint",
            bit_length=1,
            bit_offset=3,
        ),
        PacketField(
            name="CCSDS_SECONDARY_FLAG",
            data_type="uint",
            bit_length=1,
            bit_offset=4,
        ),
        PacketField(name="CCSDS_APID", data_type="uint", bit_length=11, bit_offset=5),
        PacketField(
            name="CCSDS_SEQUENCE_FLAG",
            data_type="uint",
            bit_length=2,
            bit_offset=16,
        ),
        PacketField(
            name="CCSDS_SEQUENCE_COUNT",
            data_type="uint",
            bit_length=14,
            bit_offset=18,
        ),
        PacketField(
            name="CCSDS_PACKET_LENGTH",
            data_type="uint",
            bit_length=16,
            bit_offset=32,
        ),
    ]

    return_fields.extend(existing_fields)

    return return_fields


def _get_fields_csv_file(csv_file):
    """Parse a simple comma-delimited file that defines a packet.

    Should not include the CCSDS header. The minimum set of columns are (name,
    data_type, bit_length). An optional bit_offset can also be provided.

    Parameters
    ----------
    csv_file : str
        Path to file on the local file system

    Returns
    -------
    fields : list
        A list of `PacketField` objects.
    """
    req_columns = ["name", "data_type", "bit_length"]

    with open(csv_file, "r") as fp:
        fields = []
        reader = csv.DictReader(fp, skipinitialspace=True)
        headers = reader.fieldnames

        if headers is None:
            raise RuntimeError("CSV file must not be empty")

        if not all((req_col in headers) for req_col in req_columns):
            raise ValueError(f"Minimum required columns are {req_columns}.")

        for row in reader:  # skip the header row
            if "bit_offset" not in headers:  # 3 col csv file
                if (row["data_type"].count("(") == 1) and (row["data_type"].count(")") == 1):
                    data_type = row["data_type"].split("(")[0]
                    array_shape_str = row["data_type"][
                        row["data_type"].find("(") + 1 : row["data_type"].find(")")
                    ]
                    array_shape = tuple(map(int, array_shape_str.split(", ")))
                    fields.append(
                        PacketArray(
                            name=row["name"],
                            data_type=data_type,
                            bit_length=int(row["bit_length"]),
                            array_shape=(array_shape),
                        )
                    )
                else:
                    fields.append(
                        PacketField(
                            name=row["name"],
                            data_type=row["data_type"],
                            bit_length=int(row["bit_length"]),
                        )
                    )
            if "bit_offset" in headers:  # 4 col csv file provides bit offsets
                # TODO: Check the consistency of bit_offsets versus previous bit_lengths
                if (row["data_type"].count("(") == 1) and (row["data_type"].count(")") == 1):
                    data_type = row["data_type"].split("(")[0]
                    array_shape_str = row["data_type"][
                        row["data_type"].find("(") + 1 : row["data_type"].find(")")
                    ]
                    array_shape = tuple(map(int, array_shape_str.split(", ")))
                    fields.append(
                        PacketArray(
                            name=row["name"],
                            data_type=data_type,
                            bit_length=int(row["bit_length"]),
                            array_shape=array_shape,
                        )
                    )
                else:
                    fields.append(
                        PacketField(
                            name=row["name"],
                            data_type=row["data_type"],
                            bit_length=int(row["bit_length"]),
                            bit_offset=int(row["bit_offset"]),
                        )
                    )

    return fields


def _load(file, fields, converters, decoder_name, include_primary_header=False):
    """Decode a file-like object containing a sequence of these packets.

    Parameters
    ----------
    file: str
       Path to file on the local file system, or file-like object
    fields : list of `ccsdspy.PacketField`
       Layout of packet fields contained in the definition.
    converters : dict, str to tuple (str, Converter)
       Dictionary of post-processing conversions. keys are input field names,
       values are tuples of (output_field_name, Converter instance)
    decoder_name: {'fixed_length', 'variable_length'}
       String identifying which decoder to use.
    include_primary_header: bool
       If True, provides the primary header in the output

    Returns
    -------
    dictionary mapping field names to NumPy arrays, with key order matching
    the order fields in the packet.

    Raises
    ------
    ValueError
      the decoder_name is not one of the allowed values
    """
    if hasattr(file, "read"):
        file_bytes = np.frombuffer(file.read(), "u1")
    else:
        file_bytes = np.fromfile(file, "u1")

    if include_primary_header:
        fields = _prepend_primary_header_fields(fields)

    fields, expand_history = _expand_array_fields(fields)

    if decoder_name == "fixed_length":
        field_arrays = _decode_fixed_length(file_bytes, fields)
    elif decoder_name == "variable_length":
        field_arrays = _decode_variable_length(file_bytes, fields)
    else:
        raise ValueError(
            f"Invalid decoder_name 'f{decoder_name}' specified. Must be "
            "either 'fixed_length', or 'variable_length'"
        )

    field_arrays = _unexpand_field_arrays(field_arrays, expand_history)
    field_arrays = _apply_converters(field_arrays, converters)

    return field_arrays


def _apply_converters(field_arrays, converters):
    """Apply post-processing converters in place to a dictionary of field
    arrays.

    Parameters
    ----------
    field_arrays : dict of string to NumPy arrays
       The decoded packet field arrays without any post-processing applied
    converters : dict, str to tuple (str, Converter)
       Dictionary of post-processing conversions. keys are input field names,
       values are tuples of (output_field_name, Converter instance)

    Returns
    -------
    converted_field_arrays : dict of string to NumPy arrays
       The converted decoded packet field arrays, as a dictionary with the same
       key as the passed `field_arrays`.
    """
    converted = field_arrays.copy()

    for input_field_names, (output_field_name, converter) in converters.items():
        # Collect list of input arrays to pass as *args to converter function
        input_arrays = []

        for input_field_name in input_field_names:
            input_arrays.append(field_arrays[input_field_name])

        # Call converter function
        converted[output_field_name] = converter.convert(*input_arrays)

    return converted



"""Tests for the ccsdspy.utils packate

Tests for utils.split_by_apid() can be found in test_split.py
"""
import glob
import io
import os

import numpy as np
import pytest

from .. import utils


def test_count_packets_missing_bytes():
    dir_path = os.path.dirname(os.path.realpath(__file__))
    data_path = os.path.join(dir_path, "data", "split")
    tlm_path = os.path.join(data_path, "CYGNSS_F7_L0_2022_086_10_15_V01_F__first101pkts.tlm")

    num_packets, missing_bytes = utils.count_packets(tlm_path, return_missing_bytes=True)
    assert missing_bytes == 0

    total_in_split = 0

    for apid, apid_stream in utils.split_by_apid(tlm_path).items():
        total_in_split += utils.count_packets(apid_stream)

    assert total_in_split == num_packets


def test_count_packets_simple():
    dir_path = os.path.dirname(os.path.realpath(__file__))
    data_path = os.path.join(dir_path, "data", "split")
    tlm_path = os.path.join(data_path, "CYGNSS_F7_L0_2022_086_10_15_V01_F__first101pkts.tlm")

    assert 101 == utils.count_packets(tlm_path)


def test_count_packets_file_like_obj():
    dir_path = os.path.dirname(os.path.realpath(__file__))
    data_path = os.path.join(dir_path, "data", "split")
    tlm_path = os.path.join(data_path, "CYGNSS_F7_L0_2022_086_10_15_V01_F__first101pkts.tlm")

    result_1 = utils.count_packets(tlm_path)

    with open(tlm_path, "rb") as fh:
        result_2 = utils.count_packets(fh)

    with open(tlm_path, "rb") as fh:
        result_3 = utils.count_packets(io.BytesIO(fh.read()))

    assert isinstance(result_1, int)
    assert isinstance(result_2, int)
    assert isinstance(result_3, int)

    assert result_1 == result_2
    assert result_1 == result_3


def test_split_packet_bytes_issues_warning():
    dir_path = os.path.dirname(os.path.realpath(__file__))
    data_path = os.path.join(dir_path, "data", "split")
    tlm_path = os.path.join(data_path, "CYGNSS_F7_L0_2022_086_10_15_V01_F__first101pkts.tlm")

    with open(tlm_path, "rb") as fh:
        file_bytes = fh.read()

    file_bytes = file_bytes[:-5]  # make last packet be incomplete

    with pytest.warns(UserWarning):
        result = utils.split_packet_bytes(io.BytesIO(file_bytes))


def test_split_packet_bytes_file_like_obj():
    dir_path = os.path.dirname(os.path.realpath(__file__))
    data_path = os.path.join(dir_path, "data", "split")
    tlm_path = os.path.join(data_path, "CYGNSS_F7_L0_2022_086_10_15_V01_F__first101pkts.tlm")

    result_1 = utils.split_packet_bytes(tlm_path)

    with open(tlm_path, "rb") as fh:
        result_2 = utils.split_packet_bytes(fh)

    with open(tlm_path, "rb") as fh:
        result_3 = utils.split_packet_bytes(io.BytesIO(fh.read()))

    results = [result_1, result_2, result_3]

    # check deep equality
    assert np.unique([len(res) for res in results]).size == 1

    for i in range(len(results[0])):
        for result in results:
            assert result[i] == results[0][i]


@pytest.mark.parametrize("include_primary_header", [True, False])
def test_split_packet_bytes_mixed_stream(include_primary_header):
    dir_path = os.path.dirname(os.path.realpath(__file__))
    data_path = os.path.join(dir_path, "data", "split")
    tlm_path = os.path.join(data_path, "CYGNSS_F7_L0_2022_086_10_15_V01_F__first101pkts.tlm")

    header_arrays = utils.read_primary_headers(tlm_path)

    expected_lens = header_arrays["CCSDS_PACKET_LENGTH"] + 1
    if include_primary_header:
        expected_lens += 6

    packet_bytes = utils.split_packet_bytes(tlm_path, include_primary_header=include_primary_header)

    for cur_exp_len, cur_packet_bytes in zip(expected_lens, packet_bytes):
        assert cur_exp_len == len(cur_packet_bytes)


def test_read_primary_headers():
    dir_path = os.path.dirname(os.path.realpath(__file__))
    data_path = os.path.join(dir_path, "data", "split")

    # Check files in output directory
    tlm_files = [
        "apid00384.tlm",
        "apid00386.tlm",
        "apid00391.tlm",
        "apid00392.tlm",
        "apid00393.tlm",
        "apid00394.tlm",
        "apid01313.tlm",
    ]

    for tlm_file in tlm_files:
        tlm_path = os.path.join(data_path, tlm_file)
        header_arrays = utils.read_primary_headers(tlm_path)
        apid = int(tlm_file[4:-4])

        assert np.all(header_arrays["CCSDS_VERSION_NUMBER"] == 0)
        assert np.all(header_arrays["CCSDS_PACKET_TYPE"] == 0)
        assert np.all(header_arrays["CCSDS_SECONDARY_FLAG"] == 1)
        assert np.all(header_arrays["CCSDS_SEQUENCE_FLAG"] == 3)
        assert np.all(header_arrays["CCSDS_APID"] == apid)
        assert np.all(np.diff(header_arrays["CCSDS_SEQUENCE_COUNT"]) > 0)
        assert np.unique(header_arrays["CCSDS_PACKET_LENGTH"]).size == 1  # fixed length



"""Tests for the ccsdspy.converters module"""

from datetime import datetime, timedelta
import os

import numpy as np
from numpy.testing import assert_array_equal
import pytest

from .. import converters
from .. import FixedLength, VariableLength, PacketField

TEST_FILENAME = "ccsds_converters_test.bin"
TEST_EPOCH = datetime(1970, 1, 1)


def test_converter_cannot_be_instantiated():
    with pytest.raises(NotImplementedError):
        converters.Converter()


def test_custom_class_with_missing_convert_one():
    class Test(converters.Converter):
        def __init__(self):
            pass

    with pytest.raises(NotImplementedError):
        Test().convert(np.arange(10))


@pytest.mark.parametrize("do_2darray", [False, True])
def test_poly_converter_direct_simple(do_2darray):
    coeffs = [0.3, 0.08, 5.4]

    if do_2darray:
        field_array = np.arange(200, dtype=int).reshape(5, 40)
    else:
        field_array = np.arange(10, dtype=int)

    x = field_array.astype(np.float64)
    expected = coeffs[0] * x**2 + coeffs[1] * x + coeffs[2]
    got = converters.PolyConverter(coeffs).convert(field_array)

    assert np.allclose(got, expected)


@pytest.mark.parametrize("do_2darray", [False, True])
def test_linear_converter_direct_simple(do_2darray):
    slope, intercept = 1.2, 0.4

    if do_2darray:
        field_array = np.arange(200, dtype=int).reshape(5, 40)
    else:
        field_array = np.arange(10, dtype=int)
    field_array = np.arange(10, dtype=int)

    x = field_array.astype(np.float64)
    expected = slope * x + intercept
    got = converters.LinearConverter(slope, intercept).convert(field_array)

    assert np.allclose(got, expected)


def test_enum_converter_invalid_types():
    with pytest.raises(TypeError):
        converters.EnumConverter({3.4: "fizz", 1: "buzz"})

    with pytest.raises(TypeError):
        converters.EnumConverter({0: (1, 2, 3), 1: "buzz"})

    with pytest.raises(TypeError):
        converters.EnumConverter({0: 5.2, 1: "buzz"})


@pytest.mark.parametrize("do_2darray", [False, True])
def test_enum_converter_direct_happy_path(do_2darray):
    replace_dict = {0: "OFF", 1: "ON", 2: "STANDBY", 3: "EMERGENCY"}

    field_array = np.array([0, 3, 1, 1, 3, 2])
    expected = np.array(["OFF", "EMERGENCY", "ON", "ON", "EMERGENCY", "STANDBY"])

    if do_2darray:
        shape = (3, 2)
        field_array = field_array.reshape(shape)
        expected = expected.reshape(shape)

    got = converters.EnumConverter(replace_dict).convert(field_array)

    assert np.all(got == expected)


@pytest.mark.parametrize("do_2darray", [False, True])
def test_enum_converter_direct_missing_key(do_2darray):
    replace_dict = {0: "OFF", 1: "ON", 2: "STANDBY", 3: "EMERGENCY"}

    # 8 is the missing value
    field_array = np.array([0, 4, 1, 1, 3, 2])
    expected = np.array(["OFF", "EMERGENCY", "ON", "ON", "EMERGENCY", "STANDBY"])

    if do_2darray:
        shape = (3, 2)
        field_array = field_array.reshape(shape)
        expected = expected.reshape(shape)

    with pytest.raises(converters.EnumConverterMissingKey):
        got = converters.EnumConverter(replace_dict).convert(field_array)


def test_datetime_converter_constructor_exceptions():
    with pytest.raises(TypeError):
        converters.DatetimeConverter(since="1970-01-01", units="seconds")

    with pytest.raises(TypeError):
        converters.DatetimeConverter(since=TEST_EPOCH, units=5)

    with pytest.raises(ValueError):
        converters.DatetimeConverter(since=TEST_EPOCH, units="gazooleans")

    with pytest.raises(ValueError):
        converters.DatetimeConverter(since=TEST_EPOCH, units=("gazooleans", "seconds"))


def test_date_converter_direct_one_input_days():
    field_array = np.array([100, 150, 200, 30, 499])

    # Days
    converter = converters.DatetimeConverter(since=TEST_EPOCH, units="days")
    converted = converter.convert(field_array)
    assert field_array.size == converted.size

    for i in range(converted.size):
        assert isinstance(converted[i], datetime)
        assert converted[i] == TEST_EPOCH + timedelta(days=int(field_array[i]))


def test_date_converter_direct_one_input_hours():
    field_array = np.array([100, 150, 200, 30, 499])

    # Hours
    converter = converters.DatetimeConverter(since=TEST_EPOCH, units="hours")
    converted = converter.convert(field_array)
    assert field_array.size == converted.size

    for i in range(converted.size):
        assert isinstance(converted[i], datetime)
        assert converted[i] == TEST_EPOCH + timedelta(hours=int(field_array[i]))


def test_date_converter_direct_one_input_minutes():
    field_array = np.array([100, 150, 200, 30, 499])

    # Minutes
    converter = converters.DatetimeConverter(since=TEST_EPOCH, units="minutes")
    converted = converter.convert(field_array)
    assert field_array.size == converted.size

    for i in range(converted.size):
        assert isinstance(converted[i], datetime)
        assert converted[i] == TEST_EPOCH + timedelta(minutes=int(field_array[i]))


def test_date_converter_direct_one_input_seconds():
    field_array = np.array([100, 150, 200, 30, 499])

    # Seconds
    converter = converters.DatetimeConverter(since=TEST_EPOCH, units="seconds")
    converted = converter.convert(field_array)
    assert field_array.size == converted.size

    for i in range(converted.size):
        assert isinstance(converted[i], datetime)
        assert converted[i] == TEST_EPOCH + timedelta(seconds=int(field_array[i]))


def test_date_converter_direct_one_input_milliseconds():
    field_array = np.array([100, 150, 200, 30, 499])

    # Minutes
    converter = converters.DatetimeConverter(since=TEST_EPOCH, units="milliseconds")
    converted = converter.convert(field_array)
    assert field_array.size == converted.size

    for i in range(converted.size):
        assert isinstance(converted[i], datetime)
        assert converted[i] == TEST_EPOCH + timedelta(
            seconds=int(field_array[i]) / converter._MILLISECONDS_PER_SECOND
        )


def test_date_converter_direct_one_input_microseconds():
    field_array = np.array([100, 150, 200, 30, 499])

    # Microseconds
    converter = converters.DatetimeConverter(since=TEST_EPOCH, units="microseconds")
    converted = converter.convert(field_array)
    assert field_array.size == converted.size

    for i in range(converted.size):
        assert isinstance(converted[i], datetime)
        assert converted[i] == TEST_EPOCH + timedelta(
            seconds=int(field_array[i]) / converter._MICROSECONDS_PER_SECOND
        )


def test_date_converter_direct_one_input_nanoseconds():
    field_array = np.array([100, 150, 200, 30, 499])

    # Nanoseconds
    converter = converters.DatetimeConverter(since=TEST_EPOCH, units="nanoseconds")
    converted = converter.convert(field_array)
    assert field_array.size == converted.size

    for i in range(converted.size):
        assert isinstance(converted[i], datetime)
        assert converted[i] == TEST_EPOCH + timedelta(
            seconds=int(field_array[i]) / converter._NANOSECONDS_PER_SECOND
        )


def test_date_converter_direct_multiple_inputs():
    field_array_seconds = np.array([100, 150, 200, 30, 499])
    field_array_nanoseconds = np.array([9283, 18893, 4892, 448, 2243])
    assert field_array_seconds.size == field_array_nanoseconds.size

    converter = converters.DatetimeConverter(since=TEST_EPOCH, units=("seconds", "nanoseconds"))
    converted = converter.convert(field_array_seconds, field_array_nanoseconds)
    assert field_array_seconds.size == converted.size

    for i in range(converted.size):
        assert isinstance(converted[i], datetime)

        expected = TEST_EPOCH
        expected += timedelta(seconds=int(field_array_seconds[i]))
        expected += timedelta(
            seconds=float(field_array_nanoseconds[i] / converter._NANOSECONDS_PER_SECOND)
        )
        assert converted[i] == expected


def test_stringify_bytes_invalid_format():
    with pytest.raises(ValueError):
        converters.StringifyBytesConverter(format="foobar")

    with pytest.raises(ValueError):
        converters.StringifyBytesConverter(format="")

    with pytest.raises(ValueError):
        converters.StringifyBytesConverter(format=None)


@pytest.mark.parametrize("format", ["bin", "oct", "hex"])
def test_stringify_bytes_converter_1d_uint8(format):
    field_array = np.arange(0, 60, 10, dtype=np.uint8)
    converter = converters.StringifyBytesConverter(format=format)
    result = converter.convert(field_array)

    assert isinstance(result, np.ndarray)
    assert np.issubdtype(result.dtype, object)

    if format == "bin":
        assert_array_equal(
            result,
            np.array(
                [
                    "0b0",  #   0
                    "0b1010",  #  10
                    "0b10100",  #  20
                    "0b11110",  #  30
                    "0b101000",  #  40
                    "0b110010",  #  50
                ],
                dtype=object,
            ),
        )
    elif format == "hex":
        assert_array_equal(
            result,
            np.array(
                [
                    "0x0",  #   0
                    "0xa",  #  10
                    "0x14",  #  20
                    "0x1e",  #  30
                    "0x28",  #  40
                    "0x32",  #  50
                ],
                dtype=object,
            ),
        )
    elif format == "oct":
        assert_array_equal(
            result,
            np.array(
                [
                    "0o0",  #   0
                    "0o12",  #  10
                    "0o24",  #  20
                    "0o36",  #  30
                    "0o50",  #  40
                    "0o62",  #  50
                ],
                dtype=object,
            ),
        )


@pytest.mark.parametrize("format", ["bin", "oct", "hex"])
def test_stringify_bytes_converter_2d_jagged_uint8(format):
    field_array = np.array(
        [
            np.array([0, 10, 20], dtype=np.uint8),
            np.array([], dtype=np.uint8),
            np.array([30], dtype=np.uint8),
            np.array([40, 50], dtype=np.uint8),
        ],
        dtype=object,
    )

    converter = converters.StringifyBytesConverter(format=format)
    result = converter.convert(field_array)

    assert isinstance(result, np.ndarray)
    assert np.issubdtype(result.dtype, object)

    if format == "bin":
        expected = np.array(
            [
                np.array(
                    [
                        "0b0",  #   0
                        "0b1010",  #  10
                        "0b10100",  #  20
                    ],
                    dtype=object,
                ),
                np.array([], dtype=object),
                np.array(
                    [
                        "0b11110",  #  30
                    ],
                    dtype=object,
                ),
                np.array(
                    [
                        "0b101000",  #  40
                        "0b110010",  #  50
                    ],
                    dtype=object,
                ),
            ],
            dtype=object,
        )
    elif format == "hex":
        expected = np.array(
            [
                np.array(
                    [
                        "0x0",  #   0
                        "0xa",  #  10
                        "0x14",  #  20
                    ],
                    dtype=object,
                ),
                np.array([], dtype=object),
                np.array(
                    [
                        "0x1e",  #  30
                    ],
                    dtype=object,
                ),
                np.array(
                    [
                        "0x28",  #  40
                        "0x32",  #  50
                    ],
                    dtype=object,
                ),
            ],
            dtype=object,
        )

    elif format == "oct":
        expected = np.array(
            [
                np.array(
                    [
                        "0o0",  #   0
                        "0o12",  #  10
                        "0o24",  #  20
                    ],
                    dtype=object,
                ),
                np.array([], dtype=object),
                np.array(
                    [
                        "0o36",  #  30
                    ],
                    dtype=object,
                ),
                np.array(
                    [
                        "0o50",  #  40
                        "0o62",  #  50
                    ],
                    dtype=object,
                ),
            ],
            dtype=object,
        )
    else:
        raise RuntimeError("Invalid format")

    assert expected.shape == result.shape

    for i in range(expected.shape[0]):
        assert np.issubdtype(result[i].dtype, object)
        assert_array_equal(
            result[i],
            expected[i],
        )


@pytest.mark.parametrize("format", ["bin", "oct", "hex"])
def test_stringify_bytes_converter_2d_uint16(format):
    field_array = np.arange(230, 290, 10, dtype=">u2").reshape((2, 3))
    converter = converters.StringifyBytesConverter(format=format)
    result = converter.convert(field_array)

    assert isinstance(result, np.ndarray)
    assert np.issubdtype(result.dtype, object)

    if format == "bin":
        assert_array_equal(
            result,
            np.array(
                [
                    [
                        "0b11100110",
                        "0b11110000",
                        "0b11111010",
                    ],
                    [
                        "0b100000100",
                        "0b100001110",
                        "0b100011000",
                    ],
                ],
                dtype=object,
            ),
        )
    elif format == "hex":
        assert_array_equal(
            result,
            np.array(
                [
                    [
                        "0xe6",
                        "0xf0",
                        "0xfa",
                    ],
                    [
                        "0x104",
                        "0x10e",
                        "0x118",
                    ],
                ],
                dtype=object,
            ),
        )
    elif format == "oct":
        assert_array_equal(
            result,
            np.array(
                [
                    [
                        "0o346",
                        "0o360",
                        "0o372",
                    ],
                    [
                        "0o404",
                        "0o416",
                        "0o430",
                    ],
                ],
                dtype=object,
            ),
        )


@pytest.mark.parametrize("format", ["bin", "oct", "hex"])
def test_stringify_bytes_converter_1d_uint16(format):
    field_array = np.arange(230, 290, 10, dtype=">u2")
    converter = converters.StringifyBytesConverter(format=format)
    result = converter.convert(field_array)

    assert isinstance(result, np.ndarray)
    assert np.issubdtype(result.dtype, object)

    if format == "bin":
        assert_array_equal(
            result,
            np.array(
                [
                    "0b11100110",
                    "0b11110000",
                    "0b11111010",
                    "0b100000100",
                    "0b100001110",
                    "0b100011000",
                ],
                dtype=object,
            ),
        )
    elif format == "hex":
        assert_array_equal(
            result, np.array(["0xe6", "0xf0", "0xfa", "0x104", "0x10e", "0x118"], dtype=object)
        )
    elif format == "oct":
        assert_array_equal(
            result, np.array(["0o346", "0o360", "0o372", "0o404", "0o416", "0o430"], dtype=object)
        )


def _create_simple_ccsds_packet(n=1):
    packet_version_number = int("000", 2)
    packet_type = int("0", 2)
    secondary_header_flag = int("0", 2)
    apid = 10
    sequence_flag = 1
    packet_counter = 0
    packet_data_length = 4

    CCSDS_PRIMARY_HEADER_LENGTH = 3
    total_packet_length = CCSDS_PRIMARY_HEADER_LENGTH + packet_data_length

    packet = np.zeros(total_packet_length * n, dtype=">u2")

    for i in range(n):
        packet[0 + i * total_packet_length] = (
            (packet_version_number << 13)
            + (packet_type << 12)
            + (secondary_header_flag << 11)
            + apid
        )
        packet[1 + i * total_packet_length] = (sequence_flag << 14) + packet_counter + i
        packet[2 + i * total_packet_length] = (
            packet_data_length * 2 - 1
        )  # packet length in octets minus 2

        packet[3 + i * total_packet_length] = i % 3
        packet[4 + i * total_packet_length] = i % 5
        packet[6 + i * total_packet_length] = i % 10

    packet.tofile(TEST_FILENAME)
    return packet


def test_add_converted_field_constructor_errors():
    boo_conv = converters.EnumConverter({0: "NO", 1: "YES", 2: "MAYBE"})

    pkt = VariableLength(
        [
            PacketField(name="BOO", data_type="uint", bit_length=16),
            PacketField(name="FOO", data_type="uint", bit_length=16),
            PacketField(name="BLAH", data_type="uint", bit_length=32),
        ]
    )

    # No warning
    pkt.add_converted_field("BOO", "BOO_conv", boo_conv)

    with pytest.raises(TypeError):
        pkt.add_converted_field("BOO", ("BOO_Conv",), boo_conv)

    with pytest.raises(TypeError):
        pkt.add_converted_field("BOO", ("BOO_Conv",), "BooConv")

    with pytest.raises(TypeError):
        pkt.add_converted_field({"BOO", "BIZZ"}, "BOO_Conv", boo_conv)

    with pytest.raises(ValueError):
        pkt.add_converted_field("DOES_NOT_EXIST", "BOO_Conv", boo_conv)


@pytest.mark.parametrize("cls", [FixedLength, VariableLength])
def test_end_to_end(cls):
    num_packets = 75
    packet = _create_simple_ccsds_packet(num_packets)  # noqa: F841

    coeffs = [0.52, 0.1]
    slope, intercept = 5.2, 1.2

    boo_conv = converters.EnumConverter({0: "NO", 1: "YES", 2: "MAYBE"})
    foo_conv = converters.LinearConverter(slope, intercept)
    blah_conv = converters.PolyConverter(coeffs)

    pkt = cls(
        [
            PacketField(name="BOO", data_type="uint", bit_length=16),
            PacketField(name="FOO", data_type="uint", bit_length=16),
            PacketField(name="BLAH", data_type="uint", bit_length=32),
        ]
    )

    pkt.add_converted_field("BOO", "BOO_conv", boo_conv)
    pkt.add_converted_field("FOO", "FOO_conv", foo_conv)
    pkt.add_converted_field("BLAH", "BLAH_conv", blah_conv)

    result = pkt.load(TEST_FILENAME)

    assert np.all(result["BOO"] == np.array([0, 1, 2] * (num_packets // 3)))
    assert np.all(result["FOO"] == np.arange(num_packets, dtype=int) % 5)
    assert np.allclose(result["BLAH"], np.arange(num_packets, dtype=int) % 10)
    assert np.all(result["BOO_conv"] == np.array(["NO", "YES", "MAYBE"] * (num_packets // 3)))
    assert np.all(result["FOO_conv"] == (np.arange(num_packets, dtype=int) % 5) * slope + intercept)
    assert np.allclose(
        result["BLAH_conv"], (np.arange(num_packets, dtype=int) % 10) * coeffs[0] + coeffs[1]
    )

    os.remove(TEST_FILENAME)



import numpy as np
import os

import pytest

from .. import FixedLength, VariableLength, PacketField
from ..packet_types import _inspect_primary_header_fields

TEST_FILENAME = "ccsds_primary_headers_test.bin"


def create_simple_ccsds_packet(n=1):
    packet_version_number = int("000", 2)
    packet_type = int("0", 2)
    secondary_header_flag = int("0", 2)
    apid = 10
    sequence_flag = 1
    packet_counter = 0
    packet_data_length = 4

    CCSDS_PRIMARY_HEADER_LENGTH = 3
    total_packet_length = CCSDS_PRIMARY_HEADER_LENGTH + packet_data_length

    packet = np.zeros(total_packet_length * n, dtype=">u2")

    for i in range(n):
        packet[0 + i * total_packet_length] = (
            (packet_version_number << 13)
            + (packet_type << 12)
            + (secondary_header_flag << 11)
            + apid
        )
        packet[1 + i * total_packet_length] = (sequence_flag << 14) + packet_counter + i
        packet[2 + i * total_packet_length] = (
            packet_data_length * 2 - 1
        )  # packet length in octets minus 2

        packet[3 + i * total_packet_length] = 314
        packet[4 + i * total_packet_length] = 512
        packet[6 + i * total_packet_length] = 10000

    packet.tofile(TEST_FILENAME)
    return packet


@pytest.mark.parametrize("cls", [FixedLength, VariableLength])
def test_primary_header_contents_no_offset(cls):
    """Test if the primary header is output correctly along with the data without
    defining bit offsets"""
    num_packets = 3
    packet = create_simple_ccsds_packet(num_packets)  # noqa: F841

    pkt = cls(
        [
            PacketField(name="BOO", data_type="uint", bit_length=16),
            PacketField(name="FOO", data_type="uint", bit_length=16),
            PacketField(name="BLAH", data_type="uint", bit_length=32),
        ]
    )

    result = pkt.load(TEST_FILENAME, include_primary_header=True)
    assert (result["CCSDS_VERSION_NUMBER"] == np.zeros(num_packets, dtype="uint")).all()
    assert (result["CCSDS_PACKET_TYPE"] == np.zeros(num_packets, dtype="uint")).all()
    assert (result["CCSDS_SECONDARY_FLAG"] == np.zeros(num_packets, dtype="uint")).all()
    assert (result["CCSDS_APID"] == 10 * np.ones(num_packets, dtype="uint")).all()
    assert (result["CCSDS_SEQUENCE_FLAG"] == np.ones(num_packets, dtype="uint")).all()
    assert (result["CCSDS_SEQUENCE_COUNT"] == np.arange(num_packets, dtype="uint")).all()
    assert (result["CCSDS_PACKET_LENGTH"] == 7 * np.ones(num_packets, dtype="uint")).all()
    assert (result["BOO"] == 314 * np.ones(num_packets, dtype="uint")).all()
    assert (result["FOO"] == 512 * np.ones(num_packets, dtype="uint")).all()
    assert (result["BLAH"] == 10000 * np.ones(num_packets, dtype="uint")).all()
    os.remove(TEST_FILENAME)


def test_primary_header_contents_offset():
    """Test if the primary header is output correctly along with the data with
    defining bit offsets"""
    num_packets = 3
    packet = create_simple_ccsds_packet(num_packets)  # noqa: F841

    pkt = FixedLength(
        [
            PacketField(name="BOO", data_type="uint", bit_length=16, bit_offset=48),
            PacketField(name="FOO", data_type="uint", bit_length=16, bit_offset=64),
            PacketField(name="BLAH", data_type="uint", bit_length=32, bit_offset=80),
        ]
    )

    result = pkt.load(TEST_FILENAME, include_primary_header=True)
    assert (result["CCSDS_VERSION_NUMBER"] == np.zeros(num_packets, dtype="uint")).all()
    assert (result["CCSDS_PACKET_TYPE"] == np.zeros(num_packets, dtype="uint")).all()
    assert (result["CCSDS_SECONDARY_FLAG"] == np.zeros(num_packets, dtype="uint")).all()
    assert (result["CCSDS_APID"] == 10 * np.ones(num_packets, dtype="uint")).all()
    assert (result["CCSDS_SEQUENCE_FLAG"] == np.ones(num_packets, dtype="uint")).all()
    assert (result["CCSDS_SEQUENCE_COUNT"] == np.arange(num_packets, dtype="uint")).all()
    assert (result["CCSDS_PACKET_LENGTH"] == 7 * np.ones(num_packets, dtype="uint")).all()
    assert (result["BOO"] == 314 * np.ones(num_packets, dtype="uint")).all()
    assert (result["FOO"] == 512 * np.ones(num_packets, dtype="uint")).all()
    assert (result["BLAH"] == 10000 * np.ones(num_packets, dtype="uint")).all()
    os.remove(TEST_FILENAME)


def test_primary_header_contents_offset():
    """Test if the primary header is output correctly along with the data with
    defining bit offsets"""
    num_packets = 3
    packet = create_simple_ccsds_packet(num_packets)  # noqa: F841

    pkt = FixedLength(
        [
            PacketField(name="BOO", data_type="uint", bit_length=16, bit_offset=48),
            PacketField(name="FOO", data_type="uint", bit_length=16, bit_offset=64),
            PacketField(name="BLAH", data_type="uint", bit_length=32, bit_offset=80),
        ]
    )

    result = pkt.load("ccsds_test.bin", include_primary_header=True)
    assert (result["CCSDS_VERSION_NUMBER"] == np.zeros(num_packets, dtype="uint")).all()
    assert (result["CCSDS_PACKET_TYPE"] == np.zeros(num_packets, dtype="uint")).all()
    assert (result["CCSDS_SECONDARY_FLAG"] == np.zeros(num_packets, dtype="uint")).all()
    assert (result["CCSDS_APID"] == 10 * np.ones(num_packets, dtype="uint")).all()
    assert (result["CCSDS_SEQUENCE_FLAG"] == np.ones(num_packets, dtype="uint")).all()
    assert (result["CCSDS_SEQUENCE_COUNT"] == np.arange(num_packets, dtype="uint")).all()
    assert (result["CCSDS_PACKET_LENGTH"] == 7 * np.ones(num_packets, dtype="uint")).all()
    assert (result["BOO"] == 314 * np.ones(num_packets, dtype="uint")).all()
    assert (result["FOO"] == 512 * np.ones(num_packets, dtype="uint")).all()
    assert (result["BLAH"] == 10000 * np.ones(num_packets, dtype="uint")).all()
    os.remove(TEST_FILENAME)


def test_primary_header_contents_offset():
    """Test if the primary header is output correctly along with the data with
    defining bit offsets"""
    num_packets = 3
    packet = create_simple_ccsds_packet(num_packets)  # noqa: F841

    pkt = FixedLength(
        [
            PacketField(name="BOO", data_type="uint", bit_length=16, bit_offset=48),
            PacketField(name="FOO", data_type="uint", bit_length=16, bit_offset=64),
            PacketField(name="BLAH", data_type="uint", bit_length=32, bit_offset=80),
        ]
    )

    result = pkt.load(TEST_FILENAME, include_primary_header=True)
    assert (result["CCSDS_VERSION_NUMBER"] == np.zeros(num_packets, dtype="uint")).all()
    assert (result["CCSDS_PACKET_TYPE"] == np.zeros(num_packets, dtype="uint")).all()
    assert (result["CCSDS_SECONDARY_FLAG"] == np.zeros(num_packets, dtype="uint")).all()
    assert (result["CCSDS_APID"] == 10 * np.ones(num_packets, dtype="uint")).all()
    assert (result["CCSDS_SEQUENCE_FLAG"] == np.ones(num_packets, dtype="uint")).all()
    assert (result["CCSDS_SEQUENCE_COUNT"] == np.arange(num_packets, dtype="uint")).all()
    assert (result["CCSDS_PACKET_LENGTH"] == 7 * np.ones(num_packets, dtype="uint")).all()
    assert (result["BOO"] == 314 * np.ones(num_packets, dtype="uint")).all()
    assert (result["FOO"] == 512 * np.ones(num_packets, dtype="uint")).all()
    assert (result["BLAH"] == 10000 * np.ones(num_packets, dtype="uint")).all()
    os.remove(TEST_FILENAME)


def test_check_primary_header_contents_missingseq():
    """Check that non consecutive sequence counts raises a warning."""
    num_packets = 10000
    packet_data = {
        "CCSDS_SEQUENCE_COUNT": np.arange(1, num_packets),
        "CCSDS_APID": 1 * np.ones(num_packets),
    }

    packet_data["CCSDS_SEQUENCE_COUNT"][250] = 0

    # TODO check that the warning states the right missing packets
    with pytest.warns(UserWarning):
        _inspect_primary_header_fields(packet_data)

    packet_data["CCSDS_SEQUENCE_COUNT"][500] = 0

    # TODO check that the warning states the right missing packets
    with pytest.warns(UserWarning):
        # check that it tells you that both 100 and 249 are missing
        _inspect_primary_header_fields(packet_data)


def test_check_primary_header_contents_nonconseq():
    """Check that non consecutive sequence counts raises a warning."""
    num_packets = 10000
    packet_data = {
        "CCSDS_SEQUENCE_COUNT": np.flip(np.arange(1, num_packets)),
        "CCSDS_APID": 1 * np.ones(num_packets),
    }

    with pytest.warns(UserWarning, match="out of order"):
        _inspect_primary_header_fields(packet_data)


def test_check_primary_header_contents_sameapid():
    """Check that all apids are the same."""
    num_packets = 10000
    packet_data = {
        "CCSDS_SEQUENCE_COUNT": np.arange(1, num_packets),
        "CCSDS_APID": 48 * np.ones(num_packets),
    }

    with pytest.warns(UserWarning, match="Found multiple AP IDs"):
        packet_data["CCSDS_APID"][100:200] = 58
        _inspect_primary_header_fields(packet_data)



"""This was a data cleaning script."""
import pandas as pd
from collections import OrderedDict

if __name__ == "__main__":
    df = pd.read_csv("SSAT_Mnemonics.csv")
    apids = set(df["packet_display_name"])

    for apid in apids:
        mask = df["packet_display_name"] == apid
        df_s = df[mask]

        defs = OrderedDict(
            [
                ("name", []),
                ("data_type", []),
                ("bit_offset", []),
                ("bit_length", []),
                ("calibration", []),
            ]
        )

        for i, row in df_s.iterrows():
            # Skip primary header
            if row["mnemonic"].split("_", 1)[-1].upper() in (
                "PKTVNO",
                "PCKT",
                "SHDF",
                "APID",
                "SEGF",
                "CNT",
                "PLEN",
            ):
                continue

            name = row["mnemonic"].lower()
            data_type = row["#datatype"].lower()
            bit_offset = (row["byte_offset"] - 6) * 8 + row["bit_offset"]
            bit_length = row["bit_length"]
            calibration = row["mnemonic_calibration_values"]

            if data_type == "char":
                data_type = "str"

            dim = row["dimension"]

            if dim == 1:
                defs["name"].append(name)
                defs["data_type"].append(data_type)
                defs["bit_offset"].append(bit_offset)
                defs["bit_length"].append(bit_length)
                defs["calibration"].append(calibration)
            else:
                if data_type != "float":
                    bit_length = int(bit_length / dim)

                for j in range(dim):
                    defs["name"].append(f"{name}[{j}]")
                    defs["data_type"].append(data_type)
                    defs["bit_offset"].append(bit_offset)
                    defs["bit_length"].append(bit_length)
                    defs["calibration"].append(calibration)

        out_filename = f"apid{apid:03d}/defs.csv"
        pd.DataFrame(defs).to_csv(out_filename, index=0)
        print



"""Main method of the module.

This is called with `python -m ccsdspy`
"""

__author__ = "Daniel da Silva <mail@danieldasilva.org>"

import argparse
import os
import sys

from .utils import split_by_apid


def module_main(argv=sys.argv, cwd=os.getcwd()):
    """Main method of the module, run with `python -m ccsdspy [..]"""
    # Parse command line arguments
    parser = argparse.ArgumentParser()

    subparser = parser.add_subparsers(dest="command")
    split_parser = subparser.add_parser(
        "split",
        help=(
            "Subcommand to run. Currently only split is supported. Use "
            "split to split mixed APID stream and write files to the "
            "current directory."
        ),
    )
    split_parser.add_argument("file")
    split_parser.add_argument("--valid-apids", help="Valid APIDs seperated by comma")

    args = parser.parse_args(argv[1:])

    # Implemention of split command, which splits a mixed stream and writes
    # to the current directory.
    if args.command == "split":
        if args.valid_apids:
            toks = args.valid_apids.split(",")
            valid_apids = [int(apid) for apid in toks]
        else:
            valid_apids = None

        stream_by_apid = split_by_apid(args.file, valid_apids=valid_apids)

        print("Parsing done!")

        for apid in sorted(stream_by_apid):
            if valid_apids and apid not in valid_apids:
                continue

            out_file_name = f"{cwd}/apid{apid:05d}.tlm"
            print(f"Writing {out_file_name}")

            with open(out_file_name, "wb") as file_out:
                file_out.write(stream_by_apid[apid].read())


if __name__ == "__main__":
    module_main()



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
IO Interface for Reading CCSDS Data in Python.
"""

# For egg_info test builds to pass, put package imports here.

from . import converters
from .packet_fields import PacketField, PacketArray
from .packet_types import FixedLength, VariableLength
from .utils import split_by_apid

try:
    from ._version import __version__
    from ._version import version_tuple
except ImportError:
    __version__ = "unknown"
    version_tuple = (0, 0, "unknown version")



"""High level object-oriented API for elements that can exist in a packet.

See also:
- packet_types.FixedLength
- packet_types.VariableLength
"""

__author__ = "Daniel da Silva <mail@danieldasilva.org>"

import numpy as np

from .converters import Converter


class PacketField:
    """A field contained in a packet."""

    def __init__(self, name, data_type, bit_length, bit_offset=None, byte_order="big"):
        """
        Parameters
        ----------
        name : str
            String identifier for the field. The name specified how you may
            call upon this data later.
        data_type : {'uint', 'int', 'float', 'str', 'fill'}
            Data type of the field.
        bit_length : int
            Number of bits contained in the field.
        bit_offset : int, optional
            Bit offset into packet, including the primary header which is 48 bits long.
            If this is not specified, than the bit offset will the be calculated automatically
            from its position inside the packet definition.
        byte_order : {'big', 'little'}, optional
            Byte order of the field. Defaults to big endian.

        Raises
        ------
        TypeError
             If one of the arguments is not of the correct type.
        ValueError
             data_type or byte_order is invalid
        """
        if not isinstance(name, str):
            raise TypeError("name parameter must be a str")
        if not isinstance(data_type, str):
            raise TypeError("data_type parameter must be a str")
        if not isinstance(bit_length, (int, np.integer)):
            raise TypeError("bit_length parameter must be an int")
        if not (bit_offset is None or isinstance(bit_offset, (int, np.integer))):
            raise TypeError("bit_offset parameter must be an int")

        valid_data_types = ("uint", "int", "float", "str", "fill")
        if data_type not in valid_data_types:
            raise ValueError(f"data_type must be one of {valid_data_types}")

        valid_byte_orders = ("big", "little")
        if byte_order not in valid_byte_orders:
            raise ValueError(f"byte_order must be one of {valid_byte_orders}")

        self._name = name
        self._data_type = data_type
        self._bit_length = bit_length
        self._bit_offset = bit_offset
        self._byte_order = byte_order

        self._field_type = "element"
        self._array_shape = None
        self._array_order = None

    def __repr__(self):
        values = {k: repr(v) for (k, v) in self.__dict__.items()}
        values["cls_name"] = self.__class__.__name__

        if values["cls_name"] == "PacketArray":
            return (
                "{cls_name}(name={_name}, data_type={_data_type}, "
                "bit_length={_bit_length}, bit_offset={_bit_offset}, "
                "byte_order={_byte_order}, array_shape={_array_shape}, "
                "array_order={_array_order})".format(**values)
            )
        else:
            return (
                "{cls_name}(name={_name}, data_type={_data_type}, "
                "bit_length={_bit_length}, bit_offset={_bit_offset}, "
                "byte_order={_byte_order})".format(**values)
            )

    def __iter__(self):
        return iter(
            [
                ("name", self._name),
                ("dataType", self._data_type),
                ("bitLength", self._bit_length),
                ("bitOffset", self._bit_offset),
                ("byteOrder", self._byte_order),
            ]
        )


class PacketArray(PacketField):
    """An array contained in a packet, similar to :py:class:`~ccsdspy.PacketField` but with multiple
    elements of the same size (e.g. image).
    """

    def __init__(self, *args, array_shape=None, array_order="C", **kwargs):
        """
        Parameters
        ----------
        name : str
            String identifier for the field. The name specified how you may
            call upon this data later.
        data_type : {'uint', 'int', 'float', 'str', 'fill'}
            Data type of the field.
        bit_length : int
            Number of bits contained in the field.
        array_shape : int, tuple of ints, str, 'expand'
            Shape of the array as a tuple. For a 1-dimensional array, a single integer
            can be supplied. To use another field's value, pass the name of that field. To
            grow to fill the packet, use "expand". For details on variable length fields, see
            the :py:class:`~ccsdspy.VariableLength` class.
        array_order  {'C', 'F'}
            Row-major (C-style) or column-major (Fortran-style) order.
        bit_offset : int, optional
            Bit offset into packet, including the primary header which is 48 bits long.
            If this is not specified, than the bit offset will the be calculated automatically
            from its position inside the packet definition.
        byte_order : {'big', 'little'}, optional
            Byte order of the field. Defaults to big endian.

        Raises
        ------
        TypeError
             If one of the arguments is not of the correct type.
        ValueError
             array_shape, array_order, data_type, or byte_order is invalid
        """
        if isinstance(array_shape, str):
            if "data_type" not in kwargs:
                kwargs["data_type"] = "uint"
            elif kwargs["data_type"] != "uint":
                raise ValueError("Expanding arrays must be data_type='uint'")
        else:
            if isinstance(array_shape, int):
                array_shape = (array_shape,)
            if not isinstance(array_shape, tuple):
                raise TypeError("array_shape parameter must be a tuple of ints")
            if not all(isinstance(dim, int) for dim in array_shape):
                raise TypeError("array_shape parameter must be a tuple of ints")

            if not all(dim >= 0 for dim in array_shape):
                raise TypeError("array_shape parameter dimensions must be >= 0")
            if sum(array_shape) == 0:
                raise TypeError("array must have at least one element")

        if not isinstance(array_order, str):
            raise TypeError("array_order parameter must be string")
        if array_order not in {"C", "F"}:
            raise TypeError("array_order parameter must be either 'C' or 'F'")

        super().__init__(*args, **kwargs)
        self._field_type = "array"
        self._array_shape = array_shape
        self._array_order = array_order



"""Tests for the ccsdspy.packet_fields module"""

import pytest

from .. import PacketField, PacketArray


def test_PacketField_initializer_raises_ValueError_on_bad_data_type():
    """Asserts that the PacketField class raises a ValueError when an invalid
    data_type is provided.
    """
    with pytest.raises(ValueError):
        PacketField(name="mnemonic", data_type="fizz", bit_length=1)
    with pytest.raises(ValueError):
        PacketField(name="mnemonic", data_type="uint", bit_length=1, byte_order="bloop")


def test_PacketField_initializer_raises_TypeError_on_bad_types():
    """Asserts that the PacketField class raises a TypeError
    when arguments are of the wrong type.
    """
    with pytest.raises(TypeError):
        PacketField(name=1, data_type="uint", bit_length=1)
    with pytest.raises(TypeError):
        PacketField(name="mnemonic", data_type=1, bit_length=1)
    with pytest.raises(TypeError):
        PacketField(name="mnemonic", data_type="uint", bit_length="foobar")
    with pytest.raises(TypeError):
        PacketField(name="mnemonic", data_type="uint", bit_length=4, bit_offset="foo")


def test_PacketField_repr():
    field = PacketField(name="MyField", data_type="uint", bit_length=1)

    assert "PacketField" in repr(field)
    assert "MyField" in repr(field)


def test_PacketField_iter():
    field = PacketField(name="MyField", data_type="uint", bit_length=1)
    assert dict(field)["name"] == "MyField"
    assert dict(field)["dataType"] == "uint"


def test_PacketArray_TypeErrors():
    with pytest.raises(TypeError):
        PacketArray(
            name="mnemonic",
            data_type="uint",
            array_shape=30,
            bit_length=1,
            array_order=3,  # must be str
        )

    with pytest.raises(TypeError):
        PacketArray(
            name="mnemonic",
            data_type="uint",
            array_shape=30,
            bit_length=1,
            array_order="X",  # invalid str
        )

    # Array shape must be either str or tuple
    with pytest.raises(TypeError):
        PacketArray(
            name="mnemonic",
            data_type="",
            array_shape={3, 4, 5},
            bit_length=1,
        )

    # Array shape must be >= 0 for all dims
    with pytest.raises(TypeError):
        PacketArray(
            name="mnemonic",
            data_type="",
            array_shape=(3, -2, 4),
            bit_length=1,
        )

    # Sum of array shape must be nonzero
    with pytest.raises(TypeError):
        PacketArray(
            name="mnemonic",
            data_type="",
            array_shape=(0, 0, 0),
            bit_length=1,
        )


def test_PacketArray_sets_default_data_type_expanding():
    # checks sets data type to uint when not specified (when expanding)
    field = PacketArray(name="mnemonic", bit_length=1, array_shape="expand")
    assert field._data_type == "uint"

    # checks rejects data type set thats not uint (when expanding)
    with pytest.raises(ValueError):
        PacketArray(name="mnemonic", bit_length=1, data_type="fill", array_shape="expand")



"""Internal decoding routines."""
from __future__ import division
from collections import namedtuple
import math

import numpy as np

from ccsdspy.constants import (
    BITS_PER_BYTE,
    PRIMARY_HEADER_NUM_BYTES,
)

__author__ = "Daniel da Silva <mail@danieldasilva.org>"


def _get_packet_total_bytes(primary_header_bytes):
    """Parse the number of bytes in a packet from the bytes associated
    with a packet's primary header.

    Parameters
    ----------
    primary_header_bytes : bytes
      Bytes associated with the packet primary header, of length
      `ccsdspy.constants.PRIMARY_HEADER_NUM_BYTES`.

    Returns
    -------
    num_bytes : int
      Total number of bytes in the packet, including the primary header.

    Raises
    ------
    ValueError
       The number of bytes in the supplied argument is too short. It must be
       of length `ccsdspy.constants.PRIMARY_HEADER_NUM_BYTES`.
    """
    if len(primary_header_bytes) != PRIMARY_HEADER_NUM_BYTES:
        raise ValueError(
            f"Primary header byte sequence must be {PRIMARY_HEADER_NUM_BYTES} bytes long"
        )

    # These variables are named based on 1-indexing
    primary_header_byte5 = primary_header_bytes[4]
    primary_header_byte6 = primary_header_bytes[5]

    # Number of bytes listed in the orimary header. The value in the
    # primary header is the number of byes in the body minus one.
    num_bytes = primary_header_byte5 << BITS_PER_BYTE
    num_bytes += primary_header_byte6
    num_bytes += 1
    num_bytes += PRIMARY_HEADER_NUM_BYTES

    return num_bytes


def _get_packet_apid(primary_header_bytes):
    """Parse the APID of a packet from the bytes associated
    with a packet's primary header.

    Parameters
    ----------
    primary_header_bytes : bytes
      Bytes associated with the packet primary header, of length
      `ccsdspy.constants.PRIMARY_HEADER_NUM_BYTES`.

    Raises
    ------
    ValueError
       The number of bytes in the supplied argument is too short. It must be
       of length `ccsdspy.constants.PRIMARY_HEADER_NUM_BYTES`.
    """
    if len(primary_header_bytes) != PRIMARY_HEADER_NUM_BYTES:
        raise ValueError(
            f"Primary header byte sequence must be {PRIMARY_HEADER_NUM_BYTES} bytes long"
        )

    # These variables are named based on 1-indexing
    primary_header_byte1 = primary_header_bytes[0]
    primary_header_byte2 = primary_header_bytes[1]

    # Read as 2-byte unisgned integer and mask out unwanted parts of the first
    # byte
    apid = primary_header_byte1 << BITS_PER_BYTE
    apid += primary_header_byte2
    apid &= 0x07FF

    return apid


def _decode_fixed_length(file_bytes, fields):
    """Decode a fixed length packet stream of a single APID.

    Parameters
    ----------
    file_bytes : array
       A NumPy array of uint8 type, holding the bytes of the file to decode.
    fields : list of ccsdspy.PacketField
       A list of fields, including the secondary header but excluding the
       primary header.

    Returns
    -------
    dictionary mapping field names to NumPy arrays, stored in the same order as
    the fields array passed.
    """
    # Setup a dictionary mapping a bit offset to each field. It is assumed
    # that the `fields` array contains entries for the secondary header.
    packet_nbytes = _get_packet_total_bytes(file_bytes[:PRIMARY_HEADER_NUM_BYTES])
    body_nbytes = sum(field._bit_length for field in fields) // BITS_PER_BYTE
    counter_start = max(0, (packet_nbytes - body_nbytes) * BITS_PER_BYTE)
    counter = counter_start

    bit_offset = {}

    for i, field in enumerate(fields):
        if i == 0 and field._bit_offset is not None:
            # case: using bit_offset to fix the start position
            bit_offset[field._name] = field._bit_offset
            counter = field._bit_offset + field._bit_length
        elif field._bit_offset is None:
            # case: floating start position such that packet def fills to
            # to end of packet. What's missing is assumed to be header at the beginning.
            bit_offset[field._name] = counter
            counter += field._bit_length
        elif field._bit_offset < counter:
            # case: bit_offset specifying to backtrack. This condition
            # seems odd and unlikely. Eg. one or more bits of a packet overlap?
            bit_offset[field._name] = field._bit_offset
            # don't update counter unless the the overlap goes past counter
            counter = max(field._bit_offset + field._bit_length, counter)
        elif field._bit_offset >= counter:
            # case: otherwise, bit_offset is ahead of counter and we're skipping
            # definition of 0 or more bits.
            bit_offset[field._name] = field._bit_offset
            counter = field._bit_offset + field._bit_length
        else:
            raise RuntimeError(
                f"Unexpected case: could not compare"
                f" bit_offset {field._bit_offset} with "
                f"counter {counter} for field {field._name}"
            )

    if all(field._bit_offset is None for field in fields):
        assert counter == packet_nbytes * BITS_PER_BYTE, "Field definition != packet length"
    elif counter > packet_nbytes * BITS_PER_BYTE:
        body_bits = sum(field._bit_length for field in fields)
        raise RuntimeError(
            (
                "Packet definition larger than packet length"
                f" by {counter-(packet_nbytes*BITS_PER_BYTE)} bits"
                f" (packet length in file is {packet_nbytes*BITS_PER_BYTE} bits, defined fields are {body_bits} bits)"
            )
        )

    # Setup metadata for each field, consiting of where to look for the field in
    # the file and how to parse it.
    FieldMeta = namedtuple("Meta", ["nbytes_file", "start_byte_file", "nbytes_final", "np_dtype"])
    field_meta = {}

    for field in fields:
        nbytes_file = np.ceil(field._bit_length / BITS_PER_BYTE).astype(int)
        nbytes_final = {3: 4, 5: 8, 6: 8, 7: 8}.get(nbytes_file, nbytes_file)
        start_byte_file = bit_offset[field._name] // BITS_PER_BYTE

        # byte_order_symbol is only used to control float types here.
        #  - uint and int byte order are handled with byteswap later
        #  - fill is independent of byte order (all 1's)
        #  - byte order is not applicable to str types
        byte_order_symbol = "<" if field._byte_order == "little" else ">"
        np_dtype = {
            "uint": ">u%d" % nbytes_final,
            "int": ">i%d" % nbytes_final,
            "fill": "S%d" % nbytes_final,
            "float": "%sf%d" % (byte_order_symbol, nbytes_final),
            "str": "S%d" % nbytes_final,
        }[field._data_type]

        field_meta[field] = FieldMeta(nbytes_file, start_byte_file, nbytes_final, np_dtype)

    # Read the file and calculate length of packet and number of packets in the
    # file. Trim extra bytes that may have occurred by a break in the downlink
    # while a packet was beign transferred.
    extra_bytes = file_bytes.size % packet_nbytes

    if extra_bytes > 0:
        file_bytes = file_bytes[:-extra_bytes]

    packet_count = file_bytes.size // packet_nbytes

    # Create byte arrays for each field. At the end of this method they are left
    # as the numpy uint8 type.
    field_bytes = {}

    for field in fields:
        meta = field_meta[field]
        arr = np.zeros(packet_count * meta.nbytes_final, "u1")
        xbytes = meta.nbytes_final - meta.nbytes_file

        for i in range(xbytes, meta.nbytes_final):
            arr[i :: meta.nbytes_final] = file_bytes[
                meta.start_byte_file + i - xbytes :: packet_nbytes
            ]
            field_bytes[field] = arr

    # Switch dtype of byte arrays to the final dtype, and apply masks and shifts
    # to interpret the correct bits.
    field_arrays = {}

    for field in fields:
        meta = field_meta[field]
        arr = field_bytes[field]

        if field._data_type == "int":
            # Signed integers will be treated as unsigned integers in the following
            # block, and then get special treatmenet later
            arr.dtype = meta.np_dtype.replace("i", "u")
        else:
            arr.dtype = meta.np_dtype

        if field._data_type in ("int", "uint"):
            xbytes = meta.nbytes_final - meta.nbytes_file

            bitmask_left = (
                bit_offset[field._name]
                + BITS_PER_BYTE * xbytes
                - BITS_PER_BYTE * meta.start_byte_file
            )

            bitmask_right = BITS_PER_BYTE * meta.nbytes_final - bitmask_left - field._bit_length

            bitmask_left, bitmask_right = np.array([bitmask_left, bitmask_right]).astype(
                meta.np_dtype
            )

            bitmask = np.zeros(arr.shape, arr.dtype)
            bitmask |= (1 << int(BITS_PER_BYTE * meta.nbytes_final - bitmask_left)) - 1
            tmp = np.left_shift([1], bitmask_right)
            bitmask &= np.bitwise_not(tmp[0] - 1).astype(arr.dtype)

            arr &= bitmask
            arr >>= bitmask_right

            if field._byte_order == "little":
                arr.byteswap(inplace=True)

            if field._data_type == "int":
                arr.dtype = meta.np_dtype
                sign_bit = (arr >> (field._bit_length - 1)) & 1

                # Set bits between start_bit and stop_bit to 1
                one = np.zeros_like(arr) + 1
                stop_bit = arr.itemsize * BITS_PER_BYTE
                start_bit = field._bit_length
                mask = ((one << (start_bit - one)) - one) ^ ((one << stop_bit) - one)
                arr |= sign_bit * mask

        field_arrays[field._name] = arr

    return field_arrays


def _decode_variable_length(file_bytes, fields):
    """Decode a variable length packet stream of a single APID.

    Parameters
    ----------
    file_bytes : array
       A NumPy array of uint8 type, holding the bytes of the file to decode.
    fields : list of ccsdspy.PacketField
       A list of fields, excluding the
       primary header.

    Returns
    -------
    dict
    A dictionary mapping field names to NumPy arrays, stored in the same order as the fields.
    """
    # Get start indices of each packet -------------------------------------
    packet_starts = []
    offset = 0

    while offset < len(file_bytes):
        packet_starts.append(offset)
        offset += file_bytes[offset + 4] * 256 + file_bytes[offset + 5] + 7

    assert offset == len(file_bytes)
    npackets = len(packet_starts)

    # Initialize output dicitonary of field arrays, their dtypes, and the offsets
    # that can be determined before parsing each packet.
    # ------------------------------------------------------------------------
    field_arrays, numpy_dtypes, bit_offsets = _varlength_intialize_field_arrays(fields, npackets)

    # Loop through packets
    # ----------------------------------------------------------------------------
    for pkt_num, packet_start in enumerate(packet_starts):
        packet_nbytes = file_bytes[packet_start + 4] * 256 + file_bytes[packet_start + 5] + 7
        bit_offsets_cur = bit_offsets.copy()
        bit_lengths_cur = {}

        offset_counter = 0
        offset_history = []

        for i, field in enumerate(fields):
            # Determine the bit length for field
            # ----------------------------------
            if field._array_shape == "expand":
                footer_bits = sum(field._bit_length for fld in fields[i + 1 :])
                bit_length = packet_nbytes * BITS_PER_BYTE - footer_bits - offset_counter
            elif isinstance(field._array_shape, str):
                # Defined by previous field
                bit_length = field_arrays[field._array_shape][pkt_num] * field._bit_length
            else:
                bit_length = field._bit_length

            bit_lengths_cur[field._name] = bit_length

            # Determine both offset
            if field._name not in bit_offsets_cur:
                bit_offsets_cur[field._name] = offset_counter

            offset_history.append(offset_counter)
            offset_counter += bit_length

            # Parse field data
            # ------------------
            field_raw_data = None  # will be array of uint8
            if bit_offsets_cur[field._name] < 0:
                # Footer byte after expanding field: Referenced from end of packet
                start_byte = (
                    packet_start + packet_nbytes + bit_offsets_cur[field._name] // BITS_PER_BYTE
                )
            else:
                # Header byte before expanding field: Referenced from start of packet
                start_byte = packet_start + bit_offsets_cur[field._name] // BITS_PER_BYTE

            if isinstance(field._array_shape, str):
                stop_byte = start_byte + bit_lengths_cur[field._name] // BITS_PER_BYTE
                field_raw_data = file_bytes[start_byte:stop_byte]
            else:
                # Get field_raw_data, which are the bytes of the field as uint8 for this
                # packet
                bit_offset = bit_offsets_cur[field._name]
                nbytes_file = (
                    (bit_offset + field._bit_length - 1) // BITS_PER_BYTE
                    - bit_offset // BITS_PER_BYTE
                    + 1
                )

                nbytes_final = {3: 4, 5: 8, 6: 8, 7: 8}.get(nbytes_file, nbytes_file)
                xbytes = nbytes_final - nbytes_file
                field_raw_data = np.zeros(nbytes_final, "u1")

                for i in range(xbytes, nbytes_final):
                    idx = start_byte + i - xbytes
                    field_raw_data[i] = file_bytes[idx]

            # Switch dtype of byte arrays to the final dtype, and apply masks and shifts
            # to interpret the correct bits.
            if field._data_type == "int":
                # Signed integers will be treated as unsigned integers in the following
                # block, and then get special treatmenet later
                field_raw_data.dtype = numpy_dtypes[field._name].replace("i", "u")
            else:
                field_raw_data.dtype = numpy_dtypes[field._name]

            if field._data_type in ("uint", "int"):
                if not isinstance(field._array_shape, str):
                    last_byte = start_byte + nbytes_file
                    end_last_parent_byte = last_byte * BITS_PER_BYTE

                    b = bit_offsets_cur[field._name]
                    if b < 0:
                        b = packet_nbytes * BITS_PER_BYTE + bit_offsets_cur[field._name]

                    last_occupied_bit = packet_start * BITS_PER_BYTE + b + bit_length
                    left_bits_before_shift = b % BITS_PER_BYTE
                    right_shift = end_last_parent_byte - last_occupied_bit

                    assert right_shift >= 0, f"right_shift={right_shift}, {field}"

                    if left_bits_before_shift > 0:
                        mask = int(
                            "1" * ((nbytes_file * BITS_PER_BYTE) - left_bits_before_shift), 2
                        )
                        field_raw_data &= mask

                    if right_shift > 0:
                        field_raw_data >>= right_shift

                if field._byte_order == "little":
                    field_raw_data.byteswap(inplace=True)

                if field._data_type == "int":
                    field_raw_data.dtype = numpy_dtypes[field._name]
                    sign_bit = (field_raw_data >> (field._bit_length - 1)) & 1

                    if sign_bit:
                        # Set bits between start_bit and stop_bit to 1
                        one = np.zeros_like(field_raw_data) + 1
                        stop_bit = field_raw_data.itemsize * BITS_PER_BYTE
                        start_bit = field._bit_length
                        mask = ((one << (start_bit - one)) - one) ^ ((one << stop_bit) - one)

                        field_raw_data |= mask

            # Set the field in the final array
            if isinstance(field._array_shape, str):
                field_arrays[field._name][pkt_num] = field_raw_data
            else:
                field_arrays[field._name][pkt_num] = field_raw_data[0]

    return field_arrays


def _varlength_intialize_field_arrays(fields, npackets):
    """
    Initialize output dicitonary of field arrays, their dtypes, and the offsets
    that can be determined before parsing each packet.

    Expanding fields will be an array of dtype=object (jagged array), which will be
    an array refrence at each index. Non-expanding fields are the matched to the most
    suitable data type.

    Parameters
    ----------
    fields : list of ccsdspy.PacketField
       A list of fields, including the secondary header but excluding the
       primary header.

    npackets : int
       Number of packets in the file

    Returns
    -------
    field_arrays : dict, str to array
        Dictionary of initialized field arrays, mapping string field name to numpy array
    numpy_dtypes : dict, str to numpy dtype
        Dictionary of datatypes for the final field arrays, mapping string field names
        to numpy data types
    bit_offsets : dict, str to int/None
        Dictionary of bit offsets that can be determined before parsing each packet. Maps
        string field names to integers or None
    """
    # First pass determination of bit lengths for non-variable fields
    # ---------------------------------------------------------------
    bit_offsets = {}
    counter = 0
    last_var_idx = None

    for i, field in enumerate(fields):
        if isinstance(field._array_shape, str):
            break
        elif field._bit_offset is None:
            bit_offsets[field._name] = counter
            counter += field._bit_length
        else:
            bit_offsets[field._name] = field._bit_offset
            counter = max(field._bit_offset + field._bit_length, counter)

    for i, field in enumerate(fields):
        if isinstance(field._array_shape, str):
            last_var_idx = i

    if last_var_idx is not None:
        counter = 0
        footer_fields = fields[last_var_idx + 1 :]
        for i, field in enumerate(reversed(footer_fields)):
            bit_offsets[field._name] = counter - field._bit_length
            counter -= field._bit_length

    # Generate field arrays
    # ---------------------
    field_arrays = {}
    numpy_dtypes = {}
    nbytes_file = {}

    for i, field in enumerate(fields):
        # Number of bytes that the field spans in the file
        if isinstance(field._array_shape, str):
            nbytes_file[field._name] = field._bit_length // BITS_PER_BYTE
        else:
            if i > 0 and isinstance(fields[i - 1]._array_shape, str):
                # if preceding field is vairable length, we can assume this is byte
                # asgined
                nbytes_file[field._name] = math.ceil(field._bit_length / BITS_PER_BYTE)
            else:
                bit_offset = bit_offsets[field._name]
                nbytes_file[field._name] = (
                    (bit_offset + field._bit_length - 1) // BITS_PER_BYTE
                    - bit_offset // BITS_PER_BYTE
                    + 1
                )

        # NumPy only has 2-byte, 4-byte and 8-byte variants (eg, float16, float32,
        # float64, but not float48). Map them to an nbytes for the output.
        nbytes_final = {3: 4, 5: 8, 6: 8, 7: 8}.get(
            nbytes_file[field._name], nbytes_file[field._name]
        )

        # byte_order_symbol is only used to control float types here.
        #  - uint and int byte order are handled with byteswap later
        #  - fill is independent of byte order (all 1's)
        #  - byte order is not applicable to str types
        byte_order_symbol = "<" if field._byte_order == "little" else ">"
        np_dtype = {
            "uint": ">u%d" % nbytes_final,
            "int": ">i%d" % nbytes_final,
            "fill": "S%d" % nbytes_final,
            "float": "%sf%d" % (byte_order_symbol, nbytes_final),
            "str": "S%d" % nbytes_final,
        }[field._data_type]

        numpy_dtypes[field._name] = np_dtype

        if isinstance(field._array_shape, str):
            field_arrays[field._name] = np.zeros(npackets, dtype=object)
        else:
            field_arrays[field._name] = np.zeros(npackets, dtype=np_dtype)

    return field_arrays, numpy_dtypes, bit_offsets



"""Regression Tests"""

import io
import os

import numpy as np
import pytest

from .. import FixedLength, VariableLength, PacketField, PacketArray


@pytest.mark.parametrize("pkt_class", [FixedLength, VariableLength])
def test_odd_length_neg_ints(pkt_class):
    """This fixes an issue with odd-length integers being negative (odd length
     meaning something like 24 bit).

    See: https://github.com/CCSDSPy/ccsdspy/issues/76
    """
    # This will parse the apid=1227 packet. Code obtained from @nischayn99 created
    # the definition dynamically, so repeating that here.
    TIME_SAMPLE_PER_HF_PACKET = 160

    fields = [
        PacketField(name="Instrument SCLK Time second", bit_length=32, data_type="uint"),
        PacketField(name="Instrument SCLK Time subsec", bit_length=16, data_type="uint"),
        PacketField(name="Accountability ID", bit_length=32, data_type="uint"),
    ]

    for i in range(TIME_SAMPLE_PER_HF_PACKET):
        for c in range(3, 0, -1):
            fields.append(PacketField(f"FGx_CH{c}_{i}", bit_length=24, data_type="int"))

    fields.extend(
        [
            PacketField(name="FGx_-4.7VHK", bit_length=24, data_type="int"),
            PacketField(name="FGx_+4.7VHK", bit_length=24, data_type="int"),
            PacketField(name="FGx_2VREF", bit_length=24, data_type="int"),
            PacketField(name="FGx_1VREF", bit_length=24, data_type="int"),
            PacketField(name="FGx_DRV_SNS", bit_length=24, data_type="int"),
            PacketField(name="FGx_OP_PRTA", bit_length=24, data_type="int"),
            PacketField(name="FGx_FBX", bit_length=24, data_type="int"),
            PacketField(name="FGx_FBY", bit_length=24, data_type="int"),
            PacketField(name="FGx_FBZ", bit_length=24, data_type="int"),
            PacketField(name="FGx_BPFX", bit_length=24, data_type="int"),
            PacketField(name="FGx_BPFY", bit_length=24, data_type="int"),
            PacketField(name="FGx_BPFZ", bit_length=24, data_type="int"),
            PacketField(name="FGx_+4.7_I", bit_length=24, data_type="int"),
            PacketField(name="FGx_-4.7_I", bit_length=24, data_type="int"),
            PacketField(name="FGx_HK_CH14", bit_length=24, data_type="int"),
            PacketField(name="FGx_HK_CH15", bit_length=24, data_type="int"),
            PacketField(name="Register 80", bit_length=16, data_type="uint"),
            PacketField(name="PEC (CRC-16-CCITT)", bit_length=16, data_type="uint"),
        ]
    )

    # Try parsing the packet
    dir_path = os.path.dirname(os.path.realpath(__file__))
    bin_path = os.path.join(dir_path, "data", "europa_clipper", "apid01227.tlm")

    pkt = pkt_class(fields)
    results = pkt.load(bin_path)

    assert np.all(results["FGx_CH1_134"] == -88)
    assert np.issubdtype(results["FGx_CH1_134"].dtype, np.int32)

    assert np.all(results["Accountability ID"] == 400)
    assert np.issubdtype(results["Accountability ID"].dtype, np.uint32)


@pytest.mark.parametrize("pkt_class", [FixedLength, VariableLength])
def test_nbytes_file_too_long(pkt_class):
    """This fixes an issue where nbytes_file was incorrectly incremented
    and sometimes reaches past the end of the file.

    See: https://github.com/CCSDSPy/ccsdspy/issues/78
    """

    # Create packet definition
    pkt = pkt_class(
        [
            PacketArray(
                name="twelve", data_type="int", bit_length=12, array_shape=(8, 1), array_order="C"
            )
        ]
    )

    # Test with one packets
    fakepkt = io.BytesIO(
        b"\x00\x01\xC0\x00\x00\x0B\x00\x00\x01\x00\x20\x03\x00\x40\x05\x00\x60\x07"
    )

    with open("fake_twelve_single.ccsds", "wb") as file:
        file.write(fakepkt.getbuffer())

    result = pkt.load(fakepkt)

    assert result["twelve"].shape == (1, 8, 1)
    assert np.array_equal(result["twelve"], np.arange(8).reshape((1, 8, 1)))

    # Test with two packets
    fakepkts_even = io.BytesIO(
        b"\x00\x01\xC0\x01\x00\x0B\x00\x00\x01\x00\x20\x03\x00\x40\x05\x00\x60\x07\x00\x01\xC0\x02\x00\x0B\x00\x00\x01\x00\x20\x03\x00\x40\x05\x00\x60\x07"
    )

    with open("fake_twelve_even.ccsds", "wb") as file:
        file.write(fakepkts_even.getbuffer())

    result = pkt.load(fakepkts_even)

    expected = np.array([np.arange(8), np.arange(8)]).reshape(2, 8, 1)

    assert result["twelve"].shape == (2, 8, 1)
    assert np.array_equal(result["twelve"], expected)

    # Test with three packets
    fakepkts_odd = io.BytesIO(
        b"\x00\x01\xC0\x01\x00\x0B\x00\x00\x01\x00\x20\x03\x00\x40\x05\x00\x60\x07\x00\x01\xC0\x02\x00\x0B\x00\x00\x01\x00\x20\x03\x00\x40\x05\x00\x60\x07\x00\x01\xC0\x03\x00\x0B\x00\x00\x01\x00\x20\x03\x00\x40\x05\x00\x60\x07"
    )

    with open("fake_twelve_odd.ccsds", "wb") as file:
        file.write(fakepkts_odd.getbuffer())

    result = pkt.load(fakepkts_odd)

    expected = np.array([np.arange(8), np.arange(8), np.arange(8)]).reshape(3, 8, 1)

    assert result["twelve"].shape == (3, 8, 1)
    assert np.array_equal(result["twelve"], expected)


@pytest.mark.parametrize("pkt_class", [FixedLength, VariableLength])
def test_neg_ints_flip_start_bit(pkt_class):
    """This fixes an issue where flipping the padding bits was done
    from the incorrect start bit.

    See: https://github.com/CCSDSPy/ccsdspy/issues/80
    """
    pkt = pkt_class(
        [
            PacketField(name="uinttwo", data_type="uint", bit_length=3),
            PacketField(name="negfive", data_type="int", bit_length=5),
            PacketField(name="postwelve", data_type="int", bit_length=12),
            PacketField(name="negsix", data_type="int", bit_length=12),
        ]
    )

    # 0b101, 0b11011, 0b000000001100, 0b111111110100
    fakepkt = io.BytesIO(b"\x00\x01\xC0\x00\x00\x03\x5B\x00\xCF\xFA")
    result = pkt.load(fakepkt)

    assert np.array_equal(result["uinttwo"], np.array([2], dtype=np.uint8))
    assert np.array_equal(result["negfive"], np.array([-5], dtype=np.int8))
    assert np.array_equal(result["postwelve"], np.array([12], dtype=np.int16))
    assert np.array_equal(result["negsix"], np.array([-6], dtype=np.int16))



"""This class hold the implementation of the converter system, which applies 
post-process to decoded packet fields. This post-processing includes applying
linear/polynomial calibration curves, dictionary replacement, and time parsing.
"""

from datetime import datetime, timedelta

import numpy as np

__all__ = [
    "EnumConverterMissingKey",
    "Converter",
    "PolyConverter",
    "LinearConverter",
    "EnumConverter",
    "DatetimeConverter",
    "StringifyBytesConverter",
]


class EnumConverterMissingKey(RuntimeError):
    """During conversion a value was encountered which did not have a
    corresponding key in the replacement dictionary.
    """


class Converter:
    """Base class for all converter objects.

    This class is extended to create converters, and users may extend this
    class to write their own custom converters.

    To write a converter, one must create a subclass and override either the method
    `convert(*field_arrays)`. This method implements the conversion for an entire
    sequence of decoded packet field values in a single call.
    """

    def __init__(self):
        raise NotImplementedError("This is a base class not meant to be instantiated directly")

    def convert(self, field_array):
        """Convert a sequence of decoded packet field values.

        Parameters
        ----------
        field_array : NumPy array
            decoded packet field values, must have at least one dimension

        Returns
        -------
        converted_field_array : NumPy array
            converted form of the decoded packet field values
        """
        raise NotImplementedError("This method must be overridden by a subclass")


class PolyConverter(Converter):
    """Post-processing conversion which applies calibration using a series
    of coefficients ordered from highest power to intercept.
    """

    def __init__(self, coeffs):
        """Instantiate a PolyConverter object

        Parameters
        ----------
        coeffs : list of float
           Polynomial coefficients ordered from highest power to intercept.
        """
        self._coeffs = coeffs

    def convert(self, field_array):
        """Apply the polynomial conversion.

        Parameters
        ----------
        field_array : NumPy array
            decoded packet field values, must have at least one dimension

        Returns
        -------
        converted : NumPy array
            converted form of the decoded packet field values
        """
        converted = np.zeros(field_array.shape, dtype=np.float64)

        for power, coeff in enumerate(reversed(self._coeffs)):
            converted += coeff * field_array**power

        return converted


class LinearConverter(PolyConverter):
    """Post-processing conversion which applies a linear (y=mx+b)
    transformation.
    """

    def __init__(self, slope, intercept):
        """Instantiate a LinearConverter"""
        super().__init__([slope, intercept])


class EnumConverter(Converter):
    """Post-processing conversion for applying dictionary replacement of
    integers to strings.

    If during conversion a value is encountered which does not have a
    corresponding key in the replacement dictionary, then a
    `:py:class:`~ccsdspy.converters.EnumConverterMissingKey` exception
    will be thrown.
    """

    def __init__(self, replace_dict):
        """Initialize a EnumConverter.

        Parameters
        ----------
        replace_dict : dict of int to string
           Replacement dictionary mapping integer values to string values

        Raises
        ------
        TypeError
           Either one of the keys of the replacement dictionary is not an
           integer, or one of the values is not a string.
        """
        self._replace_dict = replace_dict

        for key, value in replace_dict.items():
            if not isinstance(key, int):
                raise TypeError(
                    f"Found key in EnumConverter replacement dictionary that is "
                    f"not an integer: {repr(key)}"
                )
            if not isinstance(value, str):
                raise TypeError(
                    f"Found value in EnumConverter replacement dictionary that is "
                    f"not a string: {repr(value)}"
                )

    def convert(self, field_array):
        """Apply the enum replacement conversion.

        Parameters
        ----------
        field_array : NumPy array
            decoded packet field values, must have at least one dimension

        Returns
        -------
        converted : NumPy array
            converted form of the decoded packet field values
        """
        converted = np.zeros(field_array.shape, dtype=object)
        converted_mask = np.zeros(field_array.shape, dtype=bool)

        for key, value in self._replace_dict.items():
            converted[field_array == key] = value
            converted_mask[field_array == key] = True

        if not converted_mask.all():
            missing_keys = field_array[~converted_mask].tolist()

            raise EnumConverterMissingKey(
                f"The following were encountered which did not have "
                f"corresponding keys in the replacement dictionary: "
                f"{repr(missing_keys)}"
            )

        return converted


class DatetimeConverter(Converter):
    """Post-processing conversion for converting timestamp fields to datetime
    instances, computed using offset(s) from a reference time.

    This class supports the offsets stored in multiple input fields, for example
    where one field is a coarse time (e.g. seconds) and a second field is a fine
    time (e.g. nanoseconds). To use multiple input fields, pass a tuple of input
    field names when this converter is added to the packet.
    """

    _VALID_UNITS = (
        "days",
        "hours",
        "minutes",
        "seconds",
        "milliseconds",
        "microseconds",
        "nanoseconds",
    )
    _MILLISECONDS_PER_SECOND = 1_000
    _MICROSECONDS_PER_SECOND = 1_000_000
    _NANOSECONDS_PER_SECOND = 1_000_000_000

    def __init__(self, since, units):
        """Initialize a DatetimeConverter

        Parameters
        ----------
        since : datetime
          Reference datetime. The time stored in the field(s) is considered an
          offset to this reference. If this has timezone information attached to
          it, so will the converted datetimes.
        units : str or tuple of str
          Units string of tuples of units strings for the offset of each
          input field.  Valid units are "days", "minutes", "milliseconds",
          "microseconds", and "nanoseconds".

        Raises
        ------
        TypeError
          One of the input arguments is not of the correct type
        ValueError
          One or more of the units are invalid
        """
        if not isinstance(since, datetime):
            raise TypeError("Argument 'since' must be an instance of datetime")

        if isinstance(units, str):
            units_tuple = (units,)
        elif isinstance(units, tuple):
            units_tuple = units
        else:
            raise TypeError("Argument 'units' must be either a string or tuple")

        if not (set(units_tuple) <= set(self._VALID_UNITS)):
            raise ValueError("One or more units are invalid")

        self._since = since
        self._units = units_tuple

    def convert(self, *field_arrays):
        """Apply the datetime conversion.

        Parameters
        ----------
        field_arrays : list of NumPy array
          list of decoded packet field values, each must have at least one
          dimension

        Returns
        -------
        converted : NumPy array of object (holding datetimes)
          converted form of the decoded packet field values

        Raises
        ------
        ValueError
          Too many or too few units were provided, as compared to the
          input field arrays sent.
        """
        assert len(field_arrays) > 0, "Must have at least one input field"

        converted = []

        for field_values in zip(*field_arrays):
            converted_time = self._since

            for unit, offset_raw in zip(self._units, field_values):
                offset_raw = float(offset_raw)

                if unit == "days":
                    converted_time += timedelta(days=offset_raw)
                elif unit == "hours":
                    converted_time += timedelta(hours=offset_raw)
                elif unit == "minutes":
                    converted_time += timedelta(minutes=offset_raw)
                elif unit == "seconds":
                    converted_time += timedelta(seconds=offset_raw)
                elif unit == "milliseconds":
                    converted_time += timedelta(seconds=offset_raw / self._MILLISECONDS_PER_SECOND)
                elif unit == "microseconds":
                    converted_time += timedelta(seconds=offset_raw / self._MICROSECONDS_PER_SECOND)
                elif unit == "nanoseconds":
                    converted_time += timedelta(seconds=offset_raw / self._NANOSECONDS_PER_SECOND)

            converted.append(converted_time)

        converted = np.array(converted, dtype=object)

        return converted


class StringifyBytesConverter(Converter):
    """Post-processing conversion which converts byte arrays or multi-byte
    numbers to strings in numeric representations such as binary, hexadecimal,
    or octal.

    To convert individual bytes, the input field should be defined as a
    `~ccsdspy.PacketArray` constructed with `data_type="uint"` and
    `bit_length=8`. Otherwise, each element is converted as a single entity.

    If the field is an array, the shape of the array is retained. The strings
    generated are not padded to a fixed length.

    The converted strings contain prefixes such as `0b` (binary), `0x` (hex),
    or `0o` (octal). If the number is signed and negative, the prefixes change
    to `-0b` (binary), `-0x` (hex), or `-0o` (octal).
    """

    def __init__(self, format="hex"):
        """Instantiate a StringifyBytesConverter object

        Parameters
        ----------
        format : {"bin", "hex", "oct"}
           Format used to encode the bytes in a string.
        """
        if format not in ("bin", "hex", "oct"):
            raise ValueError(
                "The format= keyword passed to StringifyBytesConverter "
                f"must be either 'bin', 'hex', or 'oct'. Got {repr(format)}"
            )

        self._format = format

    def _stringify_number(self, num, nbytes):
        """Internal helper method to convert a number to a string.

        Parameters
        ----------
        number : int
           A single number to convert to string

        Returns
        --------
        as_string : the byte converted to a string using the format
           specified when this object was created.
        """
        if self._format == "bin":
            return bin(num)
        elif self._format == "hex":
            return hex(num)
        else:
            return oct(num)

    def convert(self, field_array):
        """Apply the conversion.

        Parameters
        ----------
        field_array : NumPy array
            decoded packet field values, must have at least two dimensions

        Returns
        -------
        converted : NumPy array
            converted form of the converted packet field values
        """
        # field_arrays may either be a 1-D array, or an N-D array where N>1
        # (this includes jagged arrays where the outer array is of
        # dtype=object). These are implemented separately.
        ndims = len(field_array.shape)

        if ndims == 1 and field_array.dtype != object:
            converted = []

            for num in field_array:
                as_string = self._stringify_number(num, field_array.itemsize)
                converted.append(as_string)
        else:
            converted = []

            for i in range(field_array.shape[0]):
                cur_array_flat = field_array[i].flatten()
                n_items = cur_array_flat.shape[0]
                cur_shape = field_array[i].shape

                # Loop over elements, converting individually
                curr_array_strings = []

                for element in cur_array_flat:
                    as_string = self._stringify_number(element, cur_array_flat.itemsize)
                    curr_array_strings.append(as_string)

                # Put back into original array shape
                curr_array_strings = np.array(curr_array_strings, dtype=object).reshape(cur_shape)
                converted.append(curr_array_strings)

        converted = np.array(converted, dtype=object)

        return converted



import csv
import glob
import os
import shutil
import tempfile

import numpy as np
import pytest

from .. import FixedLength, VariableLength, PacketField
from ..__main__ import module_main
from ..utils import split_by_apid


def test_command_line_split():
    """Tests command line interface to the code"""
    # TemporaryDirectory() will delete itself when its garbage collected
    # or when the program ends
    tmp_dir = tempfile.TemporaryDirectory()
    dir_path = os.path.dirname(os.path.realpath(__file__))
    data_path = os.path.join(dir_path, "data", "split")
    mixed_stream = os.path.join(data_path, "CYGNSS_F7_L0_2022_086_10_15_V01_F__first101pkts.tlm")
    shutil.copy(mixed_stream, tmp_dir.name)

    # Call module_main() with fake argv and custom working directory
    module_main(
        argv=[
            "__tests__",
            "split",
            mixed_stream,
        ],
        cwd=tmp_dir.name,
    )

    # Check files in output directory
    expected_files = [
        "apid00384.tlm",
        "apid00386.tlm",
        "apid00391.tlm",
        "apid00392.tlm",
        "apid00393.tlm",
        "apid00394.tlm",
        "apid01313.tlm",
    ]

    for expected_file in expected_files:
        expected_path = os.path.join(data_path, expected_file)
        got_path = os.path.join(tmp_dir.name, expected_file)

        assert os.path.exists(got_path), got_path

        with open(expected_path, "rb") as fh:
            expected_bytes = fh.read()
        with open(got_path, "rb") as fh:
            got_bytes = fh.read()

        assert expected_bytes == got_bytes


@pytest.mark.parametrize("cls", [FixedLength, VariableLength])
def test_split_by_apid_and_decode(cls):
    # Read each CSV file in the data directory
    dir_path = os.path.dirname(os.path.realpath(__file__))
    data_path = os.path.join(dir_path, "data", "split")
    csv_glob = os.path.join(data_path, "defs", "*.csv")

    csv_tables = {}

    for file_name in glob.glob(csv_glob):
        key = os.path.basename(file_name).replace(".csv", "")

        with open(file_name) as csv_file:
            csv_tables[key] = list(csv.DictReader(csv_file))

    # Create a FixedLength of VariableLength for each packet
    packet_by_apid = {}

    for row_overview in csv_tables["Overview"]:
        pkt_table = csv_tables[row_overview["Packet Short Name"]]
        apid = int(row_overview["APID_Decimal"])
        fields = []

        for row in pkt_table:
            row_mnemonic = row["Mnemonic       "]
            row_data_type = row["Type           "]
            row_data_size = int(row["Data Size"])

            if "FILL" in row_mnemonic.upper().strip():
                data_type = "fill"
            elif row_data_type[0] == "U":
                data_type = "uint"
            elif row_data_type[0] == "F":
                data_type = "float"
            elif row_data_type[0] == "I":
                data_type = "int"
            else:
                raise RuntimeError(row_data_type)

            if len(row_data_type) > 2:
                if int(row_data_type[1]) > int(row_data_type[2]):
                    byte_order = "little"
                else:
                    byte_order = "big"
            else:
                byte_order = "big"

            bit_length = row_data_size

            field = PacketField(
                name=row_mnemonic,
                data_type=data_type,
                bit_length=bit_length,
                byte_order=byte_order,
            )
            fields.append(field)

        # drop first seven because are they primary header
        packet_by_apid[apid] = cls(fields[7:])

    # Do split and test result to ground truth
    mixed_stream = os.path.join(data_path, "CYGNSS_F7_L0_2022_086_10_15_V01_F__first101pkts.tlm")
    stream_by_apid = split_by_apid(mixed_stream, valid_apids=list(packet_by_apid.keys()))

    for apid, stream_from_split in stream_by_apid.items():
        if apid in [132, 134, 389, 391]:
            # Known to have problems for acceptable reasons (insufficient packet definition)
            continue

        split_result = packet_by_apid[apid].load(
            stream_from_split,
        )

        truth_file = os.path.join(data_path, f"apid{str(apid).zfill(5)}.tlm")
        true_result = packet_by_apid[apid].load(truth_file)

        assert list(split_result.keys()) == list(true_result.keys())

        for key in split_result:
            assert np.all(split_result[key] == true_result[key])



"""Tests for VariableLength class and data in tests/data/var_length"""

import os

import numpy as np
import pytest

from .. import VariableLength, PacketField, PacketArray


@pytest.mark.parametrize("include_primary_header", [True, False])
def test_var_length_data(include_primary_header):
    dir_path = os.path.dirname(os.path.realpath(__file__))
    bin_path = os.path.join(dir_path, "data", "var_length", "var_length_packets.bin")

    pkt = VariableLength(
        [
            PacketArray(
                name="data",
                data_type="uint",
                bit_length=16,
                array_shape="expand",
            )
        ]
    )

    field_arrays = pkt.load(bin_path, include_primary_header=include_primary_header)

    assert field_arrays["data"].dtype == object
    assert field_arrays["data"].size == 10
    assert all([arr.itemsize == 2 for arr in field_arrays["data"]])

    sizes = [2, 3, 5, 7, 11, 13, 17, 19, 23, 29]

    for i in range(len(sizes)):
        try:
            assert field_arrays["data"][i].shape == (sizes[i],)
        except:
            import pdb

            pdb.set_trace()

        expected = np.arange(sizes[i], dtype="uint16") + sizes[i]
        assert np.all(field_arrays["data"][i] == expected)


@pytest.mark.parametrize("include_primary_header", [True, False])
def test_var_length_data_with_footer(include_primary_header):
    dir_path = os.path.dirname(os.path.realpath(__file__))
    bin_path = os.path.join(dir_path, "data", "var_length", "var_length_packets_with_footer.bin")

    pkt = VariableLength(
        [
            PacketArray(
                name="data",
                data_type="uint",
                bit_length=16,
                array_shape="expand",
            ),
            PacketField(name="footer", data_type="uint", bit_length=16),
        ]
    )

    field_arrays = pkt.load(bin_path, include_primary_header=include_primary_header)

    assert field_arrays["data"].dtype == object
    assert field_arrays["data"].size == 10
    assert np.all(field_arrays["footer"] == 1)

    sizes = [2, 3, 5, 7, 11, 13, 17, 19, 23, 29]

    for i in range(10):
        assert field_arrays["data"][i].shape == (sizes[i],)
        expected = np.arange(sizes[i], dtype="uint16") + sizes[i]
        assert np.all(field_arrays["data"][i] == expected)


@pytest.mark.parametrize("include_primary_header", [True, False])
def test_var_length_double_varfield_with_footer(include_primary_header):
    dir_path = os.path.dirname(os.path.realpath(__file__))
    bin_path = os.path.join(
        dir_path, "data", "var_length", "var_length_packets_double_varfield_with_footer.bin"
    )

    pkt = VariableLength(
        [
            PacketField(name="data1_len", data_type="uint", bit_length=8),
            PacketArray(
                name="data1",
                data_type="uint",
                bit_length=16,
                array_shape="data1_len",
            ),
            PacketField(name="data2_len", data_type="uint", bit_length=8),
            PacketArray(
                name="data2",
                data_type="uint",
                bit_length=16,
                array_shape="data2_len",
            ),
            PacketField(name="footer", data_type="uint", bit_length=16),
        ]
    )

    field_arrays = pkt.load(bin_path, include_primary_header=include_primary_header)

    # Check footer
    assert np.all(field_arrays["footer"] == 1)

    # Check data1 array
    assert field_arrays["data1"].dtype == object
    assert field_arrays["data1"].size == 10
    sizes = [2, 3, 5, 7, 11, 13, 17, 19, 23, 29]

    for i in range(10):
        assert field_arrays["data1"][i].shape == (sizes[i],)
        expected = np.arange(sizes[i], dtype="uint16") + sizes[i]
        assert np.all(field_arrays["data1"][i] == expected)

    # Check data2 array
    assert field_arrays["data2"].dtype == object
    assert field_arrays["data2"].size == 10
    sizes = [5, 7, 2, 8, 3, 41, 42, 1, 3, 4]

    for i in range(10):
        assert field_arrays["data2"][i].shape == (sizes[i],)
        expected = np.arange(sizes[i], dtype="uint16")
        assert np.all(field_arrays["data2"][i] == expected)
